import numpy as numpy
from colossus.halo.concentration import concentration, peaks
import warnings
warnings.filterwarnings("ignore")

__all__ = ['ConcentrationDiemerJoyce',
           'ConcentrationWDMHyperbolic',
           'ConcentrationWDMPolynomial',
           'ConcentrationPeakHeight']

class _ConcentrationCDM(object):

    def __init__(self, cosmo, scatter=True, scatter_dex=0.2, *args, **kwargs):
        """
        This class handles concentrations of the mass-concentration relation for NFW profiles
        :param lens_cosmo: an instance of LensCosmo
        """
        self._cosmo = cosmo
        self._scatter = scatter
        self._scatter_dex = scatter_dex

    def nfw_concentration(self, m, z):
        """
        Evaluates the concentration of a halo of mass 'm' at redshift z
        :param M: halo mass [M_sun]
        :param z: halo redshift
        :return:
        """
        c = self._evaluate_concentration(m, z)
        if self._scatter:
            log_c = numpy.log(c)
            c = numpy.random.lognormal(log_c, self._scatter_dex)
        return c

    def _evaluate_concentration(self, *args, **kargs):
        raise Exception(
            'Custom concentration class must have a method evaluate_concentration with inputs mass, redshift')

class _ConcentrationTurnover(object):

    def __init__(self, cdm_concentration):
        """

        :param cdm_concentration: an instantiated CDM concentration-mass relation class
        """
        self._cdm_concentration = cdm_concentration

    def nfw_concentration(self, m, z):
        """
        Evaluates the concentration of a halo of mass 'm' at redshift z
        :param M: halo mass [M_sun]
        :param z: halo redshift
        :return:
        """
        c_cdm = self._cdm_concentration.nfw_concentration(m, z)
        return c_cdm * self.suppression(m, z)

    def suppression(self, *args, **kwargs):
        raise Exception('a WDM model with a turnover must have a suppression function')

class ConcentrationDiemerJoyce(_ConcentrationCDM):

    name = 'DIEMERJOYCE19'

    def __init__(self, cosmo, scatter=True, scatter_dex=0.2, *args, **kwargs):
        """
        This class handles concentrations of the mass-concentration relation for NFW profiles
        :param cosmo: an instance of astropy cosmology
        """
        super(ConcentrationDiemerJoyce, self).__init__(cosmo, scatter, scatter_dex)

    def _evaluate_concentration(self, M, z):

        """
        Evaluates the concentration of an NFW profile

        :param M: halo mass; m200 with respect to critical density of the Universe at redshift z
        :param z: redshift
        :return: halo concentratioon
        """

        model = 'diemer19'
        mdef = '200c'
        if isinstance(M, float) or isinstance(M, int):
            M_h = M * self._cosmo.h
            c = concentration(M_h, mdef=mdef, model=model, z=z)
        else:
            if isinstance(z, numpy.ndarray) or isinstance(z, list):
                assert len(z) == len(M)
                c = []
                for (mi, zi) in zip(M, z):
                    M_h = mi * self._cosmo.h
                    c.append(concentration(M_h, mdef=mdef, model=model, z=zi))
            else:
                M_h = M * self._cosmo.h
                c = concentration(M_h, mdef=mdef, model=model, z=z)
        return c

class ConcentrationPeakHeight(_ConcentrationCDM):

    name = 'PEAK_HEIGHT_POWERLAW'

    def __init__(self, cosmo, c0, zeta, beta, scatter=True, scatter_dex=0.2):
        """
        This class handles concentrations of the mass-concentration relation for NFW profiles
        :param cosmo: an instance of astropy cosmology
        :param c0: the amplitude of the concentration-mass relation at 10^8 M_sun at z=0
        :param zeta: modifies the logarithmic slope of the concentration-mass relation (1+z)^zeta
        :param beta: the logarithmic slope of the concentration-mass relation in peak height
        :param scatter: bool; whether to include scatter in concentration-mass relation
        :param scatter_dex: scatter in concentration in dex
        """
        self._c0 = c0
        self._zeta = zeta
        self._beta = beta
        super(ConcentrationPeakHeight, self).__init__(cosmo, scatter, scatter_dex)

    def _evaluate_concentration(self, M, z):

        """
        Evaluates the concentration of an NFW profile

        :param M: halo mass; m200 with respect to critical density of the Universe at redshift z
        :param z: redshift
        :return: halo concentratioon
        """
        M_h = M * self._cosmo.h
        Mref_h = 10 ** 8 * self._cosmo.h
        nu = peaks.peakHeight(M_h, z)
        nu_ref = peaks.peakHeight(Mref_h, z)
        redshift_factor = peaks.peakHeight(M_h, 0.0) / peaks.peakHeight(M_h, z)
        c = self._c0 * (nu / nu_ref) ** -self._beta * redshift_factor ** self._zeta
        return c

class ConcentrationWDMPolynomial(_ConcentrationTurnover):

    name = 'WDM_POLYNOMIAL'

    def __init__(self, cosmo, concentration_cdm_class, log_mc, c_scale=60.0,
                 c_power=-0.17, c_power_inner=1.0, mc_suppression_redshift_evolution=True, scatter=True,
                 scatter_dex=0.2, kwargs_cdm={}):
        """

        :param cosmo: an instance of astropy cosmology
        :param concentration_cdm_class: a concentration class for CDM
        :param c_scale: the leading coefficient of the suppression term (see below)
        :param c_power: the exponent outside the parenthesis of the suppression term (see equation below)
        :param c_power_inner: the exponent inside the parenthesis of the suppression term (see equation below)
        :param mc_suppression_redshift_evolution: bool; adds the (mild) redshift evolution from Bose et al. (2016)
        :param scatter: bool; whether to include scatter in concentration-mass relation
        :param scatter_dex: scatter in concentration in dex
        :param kwargs_cdm: keyword arguments for the CDM concentration class
        """
        if 'scatter' not in kwargs_cdm.keys():
            kwargs_cdm['scatter'] = scatter
        if 'scatter_dex' not in kwargs_cdm.keys():
            kwargs_cdm['scatter_dex'] = scatter_dex
        cdm_concentration = concentration_cdm_class(cosmo, **kwargs_cdm)
        if c_power > 0:
            raise Exception('c_power parameters > 0 are unphysical')
        if c_scale < 0:
            raise Exception('c_scale parameters < 0 are unphysical')

        self._log_mc = log_mc
        self._c_scale = c_scale
        self._c_power = c_power
        self._c_power_inner = c_power_inner
        self._mc_suppression_redshift_evolution = mc_suppression_redshift_evolution
        super(ConcentrationWDMPolynomial, self).__init__(cdm_concentration)

    def suppression(self, m, z):
        """

        :param m: halo mass [same units as log_mc]
        :param z: halo redshift
        :param log_mc: mass scale where suppresion kicks in [same units as m]
        :return:
        """

        mhm = 10 ** self._log_mc
        rescale_factor = (1 + self._c_scale * (mhm / m) ** self._c_power_inner) ** self._c_power
        if self._mc_suppression_redshift_evolution:
            redshift_factor = (1 + z) ** (0.026 * z - 0.04)
        else:
            redshift_factor = 1.0
        rescale = redshift_factor * rescale_factor
        return rescale

class ConcentrationWDMHyperbolic(_ConcentrationTurnover):

    name = 'WDM_HYPERBOLIC'

    def __init__(self, cosmo, concentration_cdm_class, log_mc, a,  b, scatter=True,
                 scatter_dex=0.2, kwargs_cdm={}):
        """

        :param cosmo:
        :param concentration_cdm_class:
        :param kwargs_cdm:
        """
        if 'scatter' not in kwargs_cdm.keys():
            kwargs_cdm['scatter'] = scatter
        if 'scatter_dex' not in kwargs_cdm.keys():
            kwargs_cdm['scatter_dex'] = scatter_dex
        cdm_concentration = concentration_cdm_class(cosmo, **kwargs_cdm)
        self._a = a
        self._b = b
        self._log_mc = log_mc
        super(ConcentrationWDMHyperbolic, self).__init__(cdm_concentration)

    def suppression(self, m, z):
        """

        :param m:
        :param z:
        :param log_mc:
        :param a:
        :param b:
        :return:
        """
        mhm = 10 ** self._log_mc
        log10u = numpy.log10(m / mhm)
        argument = (log10u - self._a) / (2 * self._b)
        return 0.5 * (1 + numpy.tanh(argument))




import numpy as np
from copy import deepcopy
from pyHalo.Rendering.rendering_class_base import RenderingClassBase

class Subhalos(RenderingClassBase):

    """
    This class generates subhalos, objects that have been accreted onto the host halo of the main deflector.
    """

    def __init__(self, mass_function_model, kwargs_mass_function, spatial_distribution_model,
                 geometry, lens_cosmo):
        """

        :param mass_function_model:
        :param kwargs_mass_function:
        :param spatial_distribution_model:
        :param geometry:
        :param lens_cosmo:
        """
        kwargs_mass_function_internal = deepcopy(kwargs_mass_function)
        if 'log_m_host' in kwargs_mass_function.keys():
            kwargs_mass_function_internal['host_m200'] = 10 ** kwargs_mass_function['log_m_host']
            del kwargs_mass_function_internal['log_m_host']
        elif 'host_m200' in kwargs_mass_function.keys():
            pass
        else:
            raise Exception('must specify the host halo mass through keyword argument host_m200 or log_m_host (base 10)'
                            'when adding subhalos!')
        super(Subhalos, self).__init__(mass_function_model, kwargs_mass_function_internal, spatial_distribution_model,
                 geometry, lens_cosmo, None, None)

    def render(self):

        """
        Generates halo masses and positions for subhalos of the main deflector host halo.
        :return: mass (in Msun), x (arcsec), y (arcsec), r3d (kpc), redshift
        """
        mfunc_model = self._get_mass_function_model()
        m = mfunc_model.draw()
        x, y, r3d_kpc = self.render_positions_at_z(len(m))
        z = np.array([self._lens_cosmo.z_lens] * len(m))
        subhalo_flag = [True] * len(m)
        return m, x, y, r3d_kpc, z, subhalo_flag

    def render_positions_at_z(self, nhalos):

        """
        :param nhalos: number of halos or objects to generate
        :return: the x, y coordinate of objects in arcsec, and a 3 dimensional coordinate in kpc
        The 3d coordinate only has a clear physical interpretation for subhalos, and is used to compute truncation raddi.
        For line of sight halos it is set to None.
        """

        x_kpc, y_kpc, r3d_kpc = self._spatial_distribution_model.draw(nhalos, self._lens_cosmo.z_lens)
        if len(x_kpc) > 0:
            kpc_per_asec = self._geometry.kpc_per_arcsec(self._lens_cosmo.z_lens)
            x_arcsec = x_kpc * kpc_per_asec ** -1
            y_arcsec = y_kpc * kpc_per_asec ** -1
            return x_arcsec, y_arcsec, r3d_kpc
        else:
            return np.array([]), np.array([]), np.array([])

    def _get_mass_function_model(self, log_mlow=None, log_mhigh=None):
        """

        :param z:
        :param delta_z:
        :param log_mlow: replaces log_mlow in kwargs_mass_function if not None
        :param log_mhigh: replaces log_mhigh in kwargs_mass_function if not None
        :return:
        """
        kwargs_model = deepcopy(self._kwargs_mass_function)
        if log_mlow is not None:
            kwargs_model['log_mlow'] = log_mlow
        if log_mhigh is not None:
            kwargs_model['log_mhigh'] = log_mhigh
        kwargs_model['log_mlow'], kwargs_model['log_mhigh'] = \
            self._redshift_dependent_mass_range(self._lens_cosmo.z_lens, kwargs_model['log_mlow'],
                                                kwargs_model['log_mhigh'])
        kpc_per_asec_zlens = self._lens_cosmo.cosmo.kpc_proper_per_asec(self._lens_cosmo.z_lens)
        kwargs_model['power_law_index'] = self._kwargs_mass_function['power_law_index'] + \
                                          kwargs_model['delta_power_law_index']
        kwargs_model['normalization'] = normalization_sigmasub(kwargs_model['sigma_sub'],
                                                               kwargs_model['host_m200'], self._lens_cosmo.z_lens,
                                                               kpc_per_asec_zlens, self._geometry.cone_opening_angle,
                                                               kwargs_model['power_law_index'], kwargs_model['m_pivot'])
        return self._mass_function_model(**kwargs_model)

    def convergence_sheet_correction(self, kappa_scale, log_mlow, log_mhigh, *args, **kwargs):
        """

        :param kappa_scale:
        :param log_mlow:
        :param log_mhigh:
        :param args:
        :param kwargs:
        :return:
        """
        mass_in_subhalos = self._get_mass_function_model(log_mlow, log_mhigh).first_moment
        if mass_in_subhalos == 0:
            return [], [], []

        area = self._geometry.angle_to_physical_area(0.5 * self._geometry.cone_opening_angle, self._lens_cosmo.z_lens)
        kappa = mass_in_subhalos / self._lens_cosmo.sigma_crit_mass(self._lens_cosmo.z_lens, area)
        negative_kappa = -1 * kappa_scale * kappa
        kwargs_out = [{'kappa': negative_kappa}]
        profile_name_out = ['CONVERGENCE']
        redshifts_out = [self._lens_cosmo.z_lens]
        return kwargs_out, profile_name_out, redshifts_out

def host_scaling_function(mhalo, z, k1 = 0.88, k2 = 1.7):
    """

    :param mhalo:
    :param z:
    :param k1:
    :param k2:
    :return:
    """
    logscaling = k1 * np.log10(mhalo / 10**13) + k2 * np.log10(z + 0.5)

    return 10 ** logscaling


def normalization_sigmasub(sigma_sub, host_m200, zlens, kpc_per_asec_zlens, cone_opening_angle, plaw_index, m_pivot):
    """

    :param sigma_sub:
    :param host_m200:
    :param zlens:
    :param kpc_per_asec_zlens:
    :param cone_opening_angle:
    :param plaw_index:
    :param m_pivot:
    :return:
    """

    host_scaling = host_scaling_function(host_m200, zlens)
    a0_per_kpc2 = sigma_sub * host_scaling
    R_kpc = kpc_per_asec_zlens * (0.5 * cone_opening_angle)
    area = np.pi * R_kpc ** 2
    m_pivot_factor = m_pivot ** -(plaw_index + 1)
    normalization = area * a0_per_kpc2 * m_pivot_factor
    return normalization



from pyHalo.preset_models import *
import pytest
import numpy as np
import numpy.testing as npt


class TestPresetModels(object):

    def test_CDM(self):

        cdm = CDM(0.5, 1.5)
        _ = cdm.lensing_quantities()
        _ = preset_model_from_name('CDM')

    def test_WDM(self):

        wdm = WDM(0.5, 1.5, 8.0)
        _ = wdm.lensing_quantities()
        _ = preset_model_from_name('WDM')

    def test_ULDM(self):

        flucs_shape = 'ring'
        flucs_args = {'angle': 0.0, 'rmin': 0.9, 'rmax': 1.1}
        uldm = ULDM(0.5, 1.5, -21, flucs_shape=flucs_shape, flucs_args=flucs_args)
        _ = uldm.lensing_quantities()
        _ = preset_model_from_name('ULDM')

    def test_SIDM_core_collapse(self):
        mass_ranges_subhalos = [[6, 8], [8, 10]]
        mass_ranges_field_halos = [[6, 8], [8, 10]]
        probabilities_subhalos = [1, 1]
        probabilities_field_halos = [1, 1]
        sidm_cc = SIDM_core_collapse(0.5, 1.5, mass_ranges_subhalos, mass_ranges_field_halos,
        probabilities_subhalos, probabilities_field_halos)
        _ = sidm_cc.lensing_quantities()
        _ = preset_model_from_name('SIDM_core_collapse')

    def test_WDM_mixed(self):
        wdm_mixed = WDM_mixed(0.5, 1.5, 8.0, 0.5)
        _ = wdm_mixed.lensing_quantities()
        _ = preset_model_from_name('WDM_mixed')

    def test_WDM_general(self):
        func = preset_model_from_name('WDMGeneral')
        wdm = func(0.5, 1.5, 7.7, 2.0)
        _ = wdm.lensing_quantities()

    def test_CDM_emulator(self):

        def emulator_input_callable(*args, **kwargs):
            subhalo_infall_masses = np.array([10**7,10**8])
            subhalo_x_kpc = np.array([1.0, 1.0])
            subhalo_y_kpc = np.array([1.0, 1.0])
            subhalo_final_bound_masses = subhalo_infall_masses / 2
            subhalo_infall_concentrations = np.array([16.0, 20.0])
            return subhalo_infall_masses, subhalo_x_kpc, subhalo_y_kpc, subhalo_final_bound_masses, subhalo_infall_concentrations

        concentrations = np.array([16.0, 20.0])
        mass_array = np.array([10 ** 7, 10 ** 8])
        kwargs_cdm = {'LOS_normalization': 0.0}
        cdm_subhalo_emulator = CDMFromEmulator(0.5, 1.5, emulator_input_callable, kwargs_cdm)
        _ = cdm_subhalo_emulator.lensing_quantities()
        for i, halo in enumerate(cdm_subhalo_emulator.halos):
            npt.assert_equal(halo.mass, mass_array[i])
            npt.assert_almost_equal(halo.x, 0.1584666, 4)
            npt.assert_almost_equal(halo.y, 0.1584666, 4)
            npt.assert_equal(halo.c, concentrations[i])

        emulator_input_array = np.empty((2, 5))
        emulator_input_array[:, 0] = mass_array
        emulator_input_array[:, 1] = np.array([1.0, 1.0])
        emulator_input_array[:, 2] = np.array([1.0, 1.0])
        emulator_input_array[:, 3] = mass_array / 2
        emulator_input_array[:, 4] = concentrations
        cdm_subhalo_emulator = CDMFromEmulator(0.5, 1.5, emulator_input_array, kwargs_cdm)
        _ = cdm_subhalo_emulator.lensing_quantities()
        for i, halo in enumerate(cdm_subhalo_emulator.halos):
            npt.assert_equal(halo.mass, mass_array[i])
            npt.assert_almost_equal(halo.x, 0.1584666, 4)
            npt.assert_almost_equal(halo.y, 0.1584666, 4)
            npt.assert_equal(halo.c, concentrations[i])

if __name__ == '__main__':
     pytest.main()



import numpy as np


class RenderingClassBase(object):

    def __init__(self, mass_function_model, kwargs_mass_function, spatial_distribution_model,
                 geometry, lens_cosmo, lens_plane_redshifts, delta_z_list):

        """

        :param mass_function_model:
        :param kwargs_mass_function:
        :param spatial_distribution_model:
        :param geometry:
        :param lens_cosmo:
        :param lens_plane_redshifts:
        :param delta_z_list:
        """
        self._mass_function_model = mass_function_model
        self._spatial_distribution_model = spatial_distribution_model
        if 'draw_poisson' not in kwargs_mass_function.keys():
            kwargs_mass_function['draw_poisson'] = True
        self._kwargs_mass_function = kwargs_mass_function
        self._geometry = geometry
        self._lens_cosmo = lens_cosmo
        self._lens_plane_redshifts = lens_plane_redshifts
        self._delta_z_list = delta_z_list

    def render_positions_at_z(self, z, nhalos):

        """
        :param z: redshift
        :param nhalos: number of halos or objects to generate
        :return: the x, y coordinate of objects in arcsec, and a 3 dimensional coordinate in kpc
        The 3d coordinate only has a clear physical interpretation for subhalos, and is used to compute truncation raddi.
        For line of sight halos it is set to None.
        """

        x_kpc, y_kpc = self._spatial_distribution_model.draw(nhalos, z)
        if len(x_kpc) > 0:
            kpc_per_asec = self._geometry.kpc_per_arcsec(z)
            x_arcsec = x_kpc * kpc_per_asec ** -1
            y_arcsec = y_kpc * kpc_per_asec ** -1
            return x_arcsec, y_arcsec
        else:
            return np.array([]), np.array([])

    @staticmethod
    def _redshift_dependent_normalization(z, normalization):
        """
        Evaluates a possibly redshift-dependent line-of-sight mass function normalization
        :param z: redshift
        :param normalization: the normalization passed to create the realization
        :return:
        """
        if callable(normalization):
            norm = normalization(z)
        else:
            norm = normalization
        return norm

    @staticmethod
    def _redshift_dependent_mass_range(z, log_mlow_object, log_mhigh_object):
        """
        Evaluates a possibly redshift-dependent minimum/maximum halo mass
        :param z: redshift
        :param log_mlow_object: either a number representing the minimum halo mass, or a callable function that returns
        log10(M_min) as a function z
        :param log_mhigh_object: either a number representing the maximum halo mass, or a callable function that returns
        log10(M_max) as a function z
        :return: the minimum and maximum halo mass (in log10)
        """

        if callable(log_mlow_object):
            log_mlow = log_mlow_object(z)
        else:
            log_mlow = log_mlow_object

        if callable(log_mhigh_object):
            log_mhigh = log_mhigh_object(z)
        else:
            log_mhigh = log_mhigh_object

        return log_mlow, log_mhigh

    def convergence_sheet_correction(self, *args, **kwargs):
        return [], [], []



import numpy.testing as npt
from lenstronomy.Cosmo.lens_cosmo import LensCosmo as LensCosmoLenstronomy
from pyHalo.Halos.HaloModels.NFW import NFWSubhhalo, NFWFieldHalo
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from astropy.cosmology import FlatLambdaCDM
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.cosmology import Cosmology
import pytest

class TestNFWHalos(object):

    def setup_method(self):

        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy_instance=astropy)
        self.zhalo = 0.5
        self.zsource = 2.0
        self.lens_cosmo = LensCosmo(self.zhalo, self.zsource, cosmo)
        self.truncation_class = None
        self.concentration_class = ConcentrationDiemerJoyce(self.lens_cosmo, scatter=False)
        self.lclenstronomy = LensCosmoLenstronomy(self.zhalo, self.zsource, astropy)

    def test_lenstronomy_params(self):

        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        kwargs_profile = {'evaluate_mc_at_zlens': False, 'c_scatter': False, 'c_scatter_dex': 0.2}
        is_subhalo = False
        nfw_field_halo = NFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                      self.truncation_class, self.concentration_class, unique_tag)

        kwargs_halo, _ = nfw_field_halo.lenstronomy_params
        id = nfw_field_halo.lenstronomy_ID
        npt.assert_string_equal('NFW', id[0])

        rho0, Rs, c, r200, M200 = self.lclenstronomy.nfw_angle2physical(kwargs_halo[0]['Rs'],
                                                                        kwargs_halo[0]['alpha_Rs'])
        npt.assert_almost_equal(M200/m, 1.0, 3)

        is_subhalo = True
        nfw_subhalo = NFWSubhhalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                  self.truncation_class, self.concentration_class, unique_tag)
        c_subhalo = nfw_subhalo.c
        c_field = nfw_field_halo.c
        npt.assert_equal(True, c_subhalo <= c_field)

    def test_z_infall(self):
        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        kwargs_profile = {'evaluate_mc_at_zlens': False, 'c_scatter': False, 'c_scatter_dex': 0.2}
        is_subhalo = True
        nfw_subhalo = NFWSubhhalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                      self.truncation_class, self.concentration_class, unique_tag)
        z_infall = nfw_subhalo.z_infall
        npt.assert_equal(True, self.zhalo <= z_infall)

if __name__ == '__main__':
    pytest.main()




import numpy as np
from copy import deepcopy
from pyHalo.Rendering.line_of_sight import LineOfSightNoSheet
from scipy.integrate import quad


class TwoHaloContribution(LineOfSightNoSheet):

    """
    This class adds correlated structure associated with the host dark matter halo. The amount of structure added is
    proportional to b * corr, where b is the halo bias as computed by Sheth and Tormen (1999) and corr is the
    matter-matter correlation function. Currently, this term is implemented as a rescaling of the background density by
    b * corr, where the product is the average value computed over 2*dz, where dz is the spacing of the redshift planes
    adjacent the redshift plane of the main deflector.
    """

    def __init__(self, mass_function_model, kwargs_mass_function, spatial_distribution_model,
                 geometry, lens_cosmo, lens_plane_redshifts, delta_z_list):

        """

        :param mass_function_model:
        :param kwargs_mass_function:
        :param spatial_distribution_model:
        :param geometry:
        :param lens_cosmo:
        :param lens_plane_redshifts:
        :param delta_z_list:
        """
        if 'host_m200' in kwargs_mass_function.keys():
            host_m200 = kwargs_mass_function['host_m200']
        elif 'log_m_host' in kwargs_mass_function.keys():
            host_m200 = 10 ** kwargs_mass_function['log_m_host']
        else:
            raise Exception('must specify the host halo mass through keyword argument host_m200 or log_m_host (base 10)'
                            'when adding the two-halo term!')
        z_eval = lens_cosmo.z_lens
        idx = np.argmin(abs(np.array(lens_plane_redshifts) - z_eval))
        delta_z = delta_z_list[idx]
        boost = two_halo_enhancement_factor(z_eval, delta_z, lens_cosmo, host_m200)
        relative_enhancement = boost - 1.
        kwargs_mass_function_scaled = deepcopy(kwargs_mass_function)
        kwargs_mass_function_scaled['LOS_normalization'] *= relative_enhancement
        self._delta_z = delta_z
        super(TwoHaloContribution, self).__init__(mass_function_model, kwargs_mass_function_scaled, spatial_distribution_model,
                 geometry, lens_cosmo, lens_plane_redshifts, delta_z_list)

    def render(self):

        """
        Generates halo masses and positions for objects along the line of sight
        (except for halos from the two-halo contribution)
        :return: mass (in Msun), x (arcsec), y (arcsec), r3d (kpc), redshift
        """

        mfunc_model = self._get_mass_function_model(self._lens_cosmo.z_lens, self._delta_z)
        masses = mfunc_model.draw()
        nhalos = len(masses)
        x, y = self.render_positions_at_z(self._lens_cosmo.z_lens, nhalos)
        redshifts = np.array([self._lens_cosmo.z_lens] * nhalos)
        subhalo_flag = [False] * nhalos
        r3d = np.array([None] * nhalos)
        return masses, x, y, r3d, redshifts, subhalo_flag

def two_halo_enhancement_factor(z_lens, z_step, lens_cosmo, overdensity_m200):
    """

    :param z_lens:
    :param lens_plane_redshifts:
    :param delta_z_list:
    :param lens_cosmo:
    :param overdensity_m200:
    :return:
    """
    rmax = lens_cosmo.cosmo.D_C_transverse(z_lens + z_step) - lens_cosmo.cosmo.D_C_transverse(z_lens)
    rmin = min(rmax, 0.5)
    mean_boost = 2 * quad(lens_cosmo.twohaloterm, rmin, rmax, args=(overdensity_m200, z_lens))[0] / (rmax - rmin)
    two_halo_boost = 1 + mean_boost
    return two_halo_boost




from pyHalo.Rendering.MassFunctions.mass_function_base import *
from pyHalo.Rendering.MassFunctions.density_peaks import *
from copy import deepcopy


def preset_mass_function_models(model_name, kwargs_model={}):
    """
    This function loads the concentration models, and implements some specific models from the literature
    :param model_name: the name of the concentration model (a string)
    :param kwargs_model: keyword arguments for the model class
    :return: the concentration model
    """
    kwargs_model_out = deepcopy(kwargs_model)
    if model_name == 'SHMF_LOVELL2020':
        # calibrated for subhalos
        # from https://arxiv.org/pdf/2003.01125.pdf
        kwargs_model_out['a_wdm'] = 4.2
        kwargs_model_out['b_wdm'] = 2.5
        kwargs_model_out['c_wdm'] = -0.2
        return WDMPowerLaw, kwargs_model_out
    elif model_name == 'LOVELL2020':
        # calibrated for field halos
        # from https://arxiv.org/pdf/2003.01125.pdf
        kwargs_model_out['a_wdm'] = 2.3
        kwargs_model_out['b_wdm'] = 0.8
        kwargs_model_out['c_wdm'] = -1.0
        return ShethTormenTurnover, kwargs_model_out
    elif model_name == 'SHMF_LOVELL2014':
        kwargs_model_out['a_wdm'] = 1.0
        kwargs_model_out['b_wdm'] = 1.0
        kwargs_model_out['c_wdm'] = -1.3
        return WDMPowerLaw, kwargs_model_out
    elif model_name == 'LOVELL2014':
        kwargs_model_out['a_wdm'] = 1.0
        kwargs_model_out['b_wdm'] = 1.0
        kwargs_model_out['c_wdm'] = -1.3
        return ShethTormenTurnover, kwargs_model_out
    elif model_name == 'SCHIVE2016':
        # https://arxiv.org/pdf/1508.04621.pdf
        # calibrated for ultra-light dark matter
        kwargs_model_out['a_wdm'] = 1.0
        kwargs_model_out['b_wdm'] = 1.1
        kwargs_model_out['c_wdm'] = -2.2
        return ShethTormenTurnover, kwargs_model_out
    elif model_name == 'SHMF_SCHIVE2016':
        # https://arxiv.org/pdf/1508.04621.pdf
        # calibrated for ultra-light dark matter
        kwargs_model_out['a_wdm'] = 1.0
        kwargs_model_out['b_wdm'] = 1.1
        kwargs_model_out['c_wdm'] = -2.2
        return WDMPowerLaw, kwargs_model_out
    elif model_name == 'POWER_LAW':
        return CDMPowerLaw, kwargs_model_out
    elif model_name == 'POWER_LAW_TURNOVER_SHMF':
        return WDMPowerLaw, kwargs_model_out
    elif model_name == 'POWER_LAW_TURNOVER':
        return ShethTormenTurnover, kwargs_model_out
    elif model_name == 'SHMF_MIXED_WDM_TURNOVER':
        return MixedWDMPowerLaw, kwargs_model_out
    elif model_name == 'MIXED_WDM_TURNOVER':
        return ShethTormenMixedWDM, kwargs_model_out
    elif model_name in ['STUCKER_SHMF', 'STUCKER']:
        if 'dlogT_dlogk' not in kwargs_model.keys():
            raise Exception('Must specify |dlogT_dlogk| (absolute value of the ' \
            'logarithmic derivative of the transfer function) when using the STUCKER model.)')
        if 'a_wdm' in kwargs_model.keys():
            raise Exception('Cannot specify a_wdm with the Stucker model.')
        if 'b_wdm' in kwargs_model.keys():
            raise Exception('Cannot specify b_wdm with the Stucker model.')
        if 'c_wdm' in kwargs_model.keys():
            raise Exception('Cannot specify c_wdm with the Stucker model.')
        from pyHalo.Rendering.MassFunctions.stucker import stucker_suppression_params
        a_wdm, b_wdm, c_wdm = stucker_suppression_params(kwargs_model_out['dlogT_dlogk'])
        kwargs_model_out['a_wdm'] = a_wdm
        kwargs_model_out['b_wdm'] = b_wdm
        kwargs_model_out['c_wdm'] = c_wdm
        del kwargs_model_out['dlogT_dlogk']
        if model_name == 'STUCKER':
            return ShethTormenTurnover, kwargs_model_out
        else:
            return WDMPowerLaw, kwargs_model_out
    else:
        raise Exception('model name '+str(model_name)+' not recognized')





import numpy as np
from copy import deepcopy
from pyHalo.Rendering.rendering_class_base import RenderingClassBase


class LineOfSightNoSheet(RenderingClassBase):
    """
    This class generates halos between the observer and source that are
    not bound to the host dark matter halo around the main deflector.
    """

    def render(self):

        """
        Generates halo masses and positions for objects along the line of sight
        (except for halos from the two-halo contribution)
        :return: mass (in Msun), x (arcsec), y (arcsec), r3d (kpc), redshift
        """

        masses = np.array([])
        x = np.array([])
        y = np.array([])
        redshifts = np.array([])

        for z, dz in zip(self._lens_plane_redshifts, self._delta_z_list):
            mfunc_model = self._get_mass_function_model(z, dz)
            m = mfunc_model.draw()
            nhalos = len(m)
            _x, _y = self.render_positions_at_z(z, nhalos)
            _z = np.array([z] * len(_x))
            masses = np.append(masses, m)
            x = np.append(x, _x)
            y = np.append(y, _y)
            redshifts = np.append(redshifts, _z)

        subhalo_flag = [False] * len(masses)
        r3d = np.array([None] * len(masses))
        return masses, x, y, r3d, redshifts, subhalo_flag

    def _get_mass_function_model(self, z, delta_z, log_mlow=None, log_mhigh=None):
        """

        :param z:
        :param delta_z:
        :param log_mlow: replaces log_mlow in kwargs_mass_function if not None
        :param log_mhigh: replaces log_mhigh in kwargs_mass_function if not None
        :return:
        """
        kwargs_model = deepcopy(self._kwargs_mass_function)
        if log_mlow is not None:
            kwargs_model['log_mlow'] = log_mlow
        if log_mhigh is not None:
            kwargs_model['log_mhigh'] = log_mhigh
        kwargs_model['log_mlow'], kwargs_model['log_mhigh'] = self._redshift_dependent_mass_range(z,
                                                                       kwargs_model['log_mlow'],
                                                                       kwargs_model['log_mhigh'])
        mfunc_model = self._mass_function_model.from_redshift(z, delta_z, self._geometry, kwargs_model)
        return mfunc_model


class LineOfSight(LineOfSightNoSheet):

    """
    This class generates halos between the observer and source that are not bound to the host dark matter halo around
    the main deflector, with the inclusion of negative sheets of convergence to remove the mass added in halos
    """

    def convergence_sheet_correction(self, kappa_scale, log_mlow, log_mhigh, zmin=None, zmax=None, *args, **kwargs):

        """
        this routine applies the negative convergence sheet correction for lens planes along the line of sight
        :param kwargs_mass_sheets: keyword arguments that overwrite whatever the default settings for the mass sheet
        sheet are - leave it as None for most applications
        :return:
        """

        lens_plane_redshifts = self._lens_plane_redshifts
        delta_zs = self._delta_z_list
        kwargs_out = []
        profile_names_out = []
        redshifts = []

        if zmin is None:
            zmin = 0.0
        if zmax is None:
            zmax = self._geometry._zsource

        for z, delta_z in zip(lens_plane_redshifts, delta_zs):

            if z < zmin:
                continue
            if z > zmax:
                continue

            log_mass_sheet_correction_min, log_mass_sheet_correction_max = self._redshift_dependent_mass_range(
                z, log_mlow, log_mhigh)
            mass_function_model = self._get_mass_function_model(z, delta_z,
                                                                log_mlow=log_mass_sheet_correction_min,
                                                                log_mhigh=log_mass_sheet_correction_max)
            first_moment = mass_function_model.first_moment
            kappa = self._convergence_at_z(z, self._geometry, self._lens_cosmo, first_moment)
            if kappa > 0:
                kwargs_out.append({'kappa': - kappa_scale * kappa})
                profile_names_out += ['CONVERGENCE']
                redshifts.append(z)
        return kwargs_out, profile_names_out, redshifts

    @staticmethod
    def _convergence_at_z(z, geometry_class, lens_cosmo_class, mtheory):

        area = geometry_class.angle_to_physical_area(0.5 * geometry_class.cone_opening_angle, z)
        sigma_crit_mass = lens_cosmo_class.sigma_crit_mass(z, area)
        return mtheory / sigma_crit_mass



from pyHalo.Halos.halo_base import Halo
from lenstronomy.LensModel.Profiles.general_nfw import GNFW
import numpy as np

class GeneralNFWSubhalo(Halo):

    """
    The base class for a halo with a logarithmic inner slope gamma_inner and a logarithmic outer profile
    slope gamma_outer. The normalization is defined in terms of a parameter x_match, defined as the multiple of
    rs where this profile enclodes the same mass as an NFW profile. The scale radius is assume to be the same as that of
    an NFW profile with the specified halo mass
    """
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args, truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._prof = GNFW()
        self._lens_cosmo = lens_cosmo_instance
        self._truncation_class = truncation_class
        self._concentration_class = concentration_class
        mdef = 'GNFW'
        super(GeneralNFWSubhalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                              lens_cosmo_instance, args, unique_tag)

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_lenstronomy_args'):

            (concentration, gamma_inner, gamma_outer) = self.profile_args
            rhos, rs, r200 = self._lens_cosmo.NFW_params_physical(self.mass, concentration, self.z)
            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)

            if 'x_match' in self._args.keys():
                if self._args['x_match'] == 'c':
                    x_match = concentration
                else:
                    x_match = self._args['x_match']
            else:
                # r_vmax = 2.16 * rs
                x_match = 2.16

            rs_arcsec = rs / kpc_per_arcsec
            r_match_arcsec = x_match * rs / kpc_per_arcsec
            fx = np.log(1 + x_match) - x_match / (1 + x_match)
            m_nfw = 4 * np.pi * rs ** 3 * rhos * fx

            sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
            sigma_crit_arcsec = sigma_crit_mpc * (0.001 * kpc_per_arcsec) ** 2

            rho0 = m_nfw / self._prof.mass_3d(r_match_arcsec, rs_arcsec, sigma_crit_arcsec, gamma_inner, gamma_outer)
            alpha_Rs = self._prof.rho02alpha(rho0, rs_arcsec, gamma_inner, gamma_outer)

            x, y = np.round(self.x, 4), np.round(self.y, 4)
            rs_arcsec = np.round(rs_arcsec, 10)
            alpha_Rs = np.round(alpha_Rs, 10)

            self._lenstronomy_args = [{'alpha_Rs': alpha_Rs, 'Rs': rs_arcsec, 'gamma_inner': gamma_inner, 'center_x': x, 'center_y': y,
                                      'gamma_outer': gamma_outer}]

        return self._lenstronomy_args, None

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['GNFW']

    @property
    def z_eval(self):
        """
        Returns the redshift at which to evalate the concentration-mass relation
        """
        if not hasattr(self, '_zeval'):

            if 'evaluate_mc_at_zlens' in self._args.keys() and self._args['evaluate_mc_at_zlens']:
                self._zeval = self.z
            else:
                self._zeval = self.z_infall

        return self._zeval

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):

            concentration = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
            gamma_inner = self._args['gamma_inner']
            gamma_outer = self._args['gamma_outer']
            self._profile_args = (concentration, gamma_inner, gamma_outer)

        return self._profile_args


class GeneralNFWFieldHalo(GeneralNFWSubhalo):
    """
    Class that defines a power law halo in the field
    """
    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):

            concentration = self._concentration_class.nfw_concentration(self.mass, self.z)
            gamma_inner = self._args['gamma_inner']
            gamma_outer = self._args['gamma_outer']
            self._profile_args = (concentration, gamma_inner, gamma_outer)

        return self._profile_args



import numpy
from scipy.interpolate import interp1d
from scipy.optimize import minimize
from scipy.special import erfc
from lenstronomy.Cosmo.nfw_param import NFWParam
import astropy.units as un
from colossus.lss.bias import twoHaloTerm
from scipy.integrate import quad

class LensCosmo(object):

    def __init__(self, z_lens=None, z_source=None, cosmology=None):

        if cosmology is None:
            from pyHalo.Cosmology.cosmology import Cosmology
            cosmology = Cosmology()

        self.cosmo = cosmology
        self._arcsec = 2 * numpy.pi / 360 / 3600
        self.h = self.cosmo.h
        # critical density of the universe in M_sun h^2 Mpc^-3
        rhoc = un.Quantity(self.cosmo.astropy.critical_density(0), unit=un.Msun / un.Mpc ** 3).value
        self.rhoc = rhoc / self.cosmo.h ** 2

        if z_lens is not None and z_source is not None:
            # critical density for lensing in units M_sun * Mpc ^ -2
            self.sigma_crit_lensing = self.get_sigma_crit_lensing(z_lens, z_source)
            # critical density for lensing in units M_sun * kpc ^ -2
            self.sigma_crit_lens_kpc = self.sigma_crit_lensing * (0.001) ** 2
            # critical density for lensing in units M_sun * arcsec ^ -2 at lens redshift
            self.sigmacrit = self.sigma_crit_lensing * (0.001) ** 2 * self.cosmo.kpc_proper_per_asec(z_lens) ** 2
            # lensing distances
            self.D_d, self.D_s, self.D_ds = self.cosmo.D_A_z(z_lens), self.cosmo.D_A_z(z_source), self.cosmo.D_A(
                z_lens, z_source)
        self._computed_zacc_pdf = False
        self._nfw_param = NFWParam(self.cosmo.astropy)
        self.z_lens = z_lens
        self.z_source = z_source

    def two_halo_boost(self, m200, z, rmin=0.5, rmax=10):

        """
        Computes the average contribution of the two halo term in a redshift slice adjacent
        the main deflector. Returns a rescaling factor applied to the mass function normalization

        :param m200: host halo mass
        :param z: redshift
        :param rmin: lower limit of the integral, something like the virial radius ~500 kpc
        :param rmax: Upper limit of the integral, this is computed based on redshift spacing during
        the rendering of halos
        :return: scaling factor applied to the normalization of the LOS mass function
        """

        mean_boost = 2 * quad(self.twohaloterm, rmin, rmax, args=(m200, z))[0] / (rmax - rmin)
        # factor of two for symmetry in front/behind host halo

        return 1. + mean_boost

    def twohaloterm(self, r, M, z, mdef='200c'):

        """
        Computes the boost to the background density of the Universe
        from correlated structure around a host of mass M
        :param r:
        :param M:
        :param z:
        :param mdef:
        :return:
        """

        h = self.cosmo.h
        M_h = M * h
        r_h = r * h
        rho_2h = twoHaloTerm(r_h, M_h, z, mdef=mdef) / self.cosmo._colossus_cosmo.rho_m(z)
        return rho_2h

    def nfw_physical2angle(self, m, c, z):
        """
        converts the physical mass and concentration parameter of an NFW profile into the lensing quantities
        updates lenstronomy implementation with arbitrary redshift

        :param m: mass enclosed 200 rho_crit in units of M_sun (physical units, meaning no little h)
        :param c: NFW concentration parameter (r200/r_s)
        :return: Rs_angle (angle at scale radius) (in units of arcsec), alpha_Rs (observed bending angle at the scale radius
        """
        dd = self.cosmo.D_A_z(z)
        rho0, Rs, r200 = self.nfwParam_physical(m, c, z)
        Rs_angle = Rs / dd / self._arcsec  # Rs in arcsec
        alpha_Rs = rho0 * (4 * Rs ** 2 * (1 + numpy.log(1. / 2.)))
        sigma_crit = self.get_sigma_crit_lensing(z, self.z_source)
        return Rs_angle, alpha_Rs / sigma_crit / dd / self._arcsec

    def rN_M(self, M, z, N):
        """
        computes the radius R_N of a halo of mass M in physical mass M/h, where N is a number multiplying the critical
        density of the Universe at z

        :param M: halo mass in M_sun/h
        :param z: redshift
        :param N: number, e.g. N=200 computes r200
        :return: radius R_N in physical Mpc/h
        """
        rn_mpc_over_h = (3 * M / (4 * numpy.pi * self._nfw_param.rhoc_z(z) * N)) ** (1. / 3.)
        return rn_mpc_over_h / self.cosmo.h

    def nfwParam_physical(self, m, c, z):
        """
        returns the NFW parameters in physical units
        updates lenstronomy implementation with arbitrary redshift

        :param m: physical mass in M_sun
        :param c: concentration
        :return: rho0 [Msun/Mpc^3], Rs [Mpc], r200 [Mpc]
        """
        r200 = self._nfw_param.r200_M(m * self.h, z) / self.h  # physical radius r200
        rho0 = self._nfw_param.rho0_c(c, z) * self.h**2  # physical density in M_sun/Mpc**3
        Rs = r200/c
        return rho0, Rs, r200

    def NFW_params_physical(self, m, c, z):
        """
        returns the NFW parameters in physical units

        :param M: physical mass in M_sun
        :param c: concentration
        :return: rho0 [Msun/kpc^3], Rs [kpc], r200 [kpc]
        """
        rho0, Rs, r200 = self.nfwParam_physical(m, c, z)
        return rho0 * 1000 ** -3, Rs * 1000, r200 * 1000

    def sigma_crit_mass(self, z, area):

        """
        :param z: redshift
        :param area: physical area in Mpc^2
        :return: the 'critical density mass' sigma_crit * A in units M_sun
        """

        sigma_crit_mpc = self.get_sigma_crit_lensing(z, self.z_source)
        return area * sigma_crit_mpc

    @property
    def colossus(self):
        return self.cosmo.colossus

    ######################################################
    """ACCESS ROUTINES IN STRUCTURAL PARAMETERS CLASS"""
    ######################################################

    def mthermal_to_halfmode(self, m_thermal):
        """
        Convert thermal relic particle mass to half-mode mass
        :param m_thermaal:
        :return:
        """

        # too lazy for algebra
        def _func(m):
            return abs(self.halfmode_to_thermal(m)-m_thermal)/0.01
        return minimize(_func, x0=10**8, method='Nelder-Mead')['x']

    def halfmode_to_thermal(self, m_half_mode):

        """
        Converts a half mode mass in units of solar masses (no little h) to the mass of
        the corresponding thermal relic particle in keV
        :param m: half mode mass in solar masses
        :return: thermal relic particle mass in keV
        """

        omega_matter = self.cosmo.astropy.Om0
        return 2.32 * (omega_matter / 0.25)**0.4 * (self.cosmo.h/0.7)**0.8 * \
               (m_half_mode / 10 ** 9) ** (-0.3)

    def mhm_to_fsl(self, m_hm):
        """
        Converts half mode mass to free streaming length in Mpc
        See Equations 5-8 in https://arxiv.org/pdf/1112.0330.pdf
        :param m_hm: half-mode mass in units M_sun (no little h)
        :return: free streaming length in units Mpc
        """

        rhoc = self.rhoc * self.cosmo.h ** 2
        l_hm = 2 * (3 * m_hm / (4 * numpy.pi * rhoc)) ** (1. / 3)
        l_fs = l_hm / 13.93
        return l_fs

    ##################################################################################
    """ROUTINES RELATED TO LENSING STUFF"""
    ##################################################################################

    def get_sigma_crit_lensing(self, z1, z2):

        """

        :param z1: redshift lens
        :param z2: redshift source
        :return: critical density for lensing in units of M_sun / Mpc ^ 2
        """
        D_ds = self.cosmo.D_A(z1, z2)
        D_d = self.cosmo.D_A_z(z1)
        D_s = self.cosmo.D_A_z(z2)
        d_inv = D_s*D_ds**-1*D_d**-1
        # (Mpc ^2 / sec^2) * (Mpc^-3 M_sun^1 sec ^ 2) * Mpc ^-1 = M_sun / Mpc ^2
        epsilon_crit = (self.cosmo.c**2*(4*numpy.pi*self.cosmo.G)**-1)*d_inv
        return epsilon_crit

    # ##################################################################################
    # """Routines relevant for NFW profiles"""
    # ##################################################################################
    # def NFW_params_physical(self, M, c, z):
    #     """
    #
    #     :param M: physical M200
    #     :param c: concentration
    #     :param z: halo redshift
    #     :return: physical NFW parameters in kpc units
    #     """
    #
    #     rho0, Rs, r200 = self.nfwParam_physical_Mpc(M, c, z)
    #     return rho0 * 1000 ** -3, Rs * 1000, r200 * 1000
    #
    # def nfw_physical2angle_fromNFWparams(self, rhos, rs, z):
    #
    #     """
    #     computes the deflection angle properties of an NFW halo from the density normalization mass and scale radius
    #     :param rhos: central density normalization in M_sun / Mpc^3
    #     :param rs: scale radius in Mpc
    #     :param z: redshift
    #     :return: theta_Rs (deflection angle at the scale radius) [arcsec], scale radius [arcsec]
    #     """
    #
    #     D_d = self.cosmo.D_A_z(z)
    #     Rs_angle = rs / D_d / self.cosmo.arcsec  # Rs in arcsec
    #     theta_Rs = rhos * (4 * rs ** 2 * (1 + numpy.log(1. / 2.)))
    #     eps_crit = self.get_sigma_crit_lensing(z, self.z_source)
    #
    #     return Rs_angle, theta_Rs / eps_crit / D_d / self.cosmo.arcsec
    #
    # def nfw_physical2angle(self, M, c, z):
    #     """
    #     converts the physical mass and concentration parameter of an NFW profile into the lensing quantities
    #     :param M: mass enclosed 200 \rho_crit
    #     :param c: NFW concentration parameter (r200/r_s)
    #     :return: theta_Rs (observed bending angle at the scale radius, Rs_angle (angle at scale radius) (in units of arcsec)
    #     """
    #
    #     rhos, rs, _ = self.nfwParam_physical_Mpc(M, c, z)
    #     return self.nfw_physical2angle_fromNFWparams(rhos, rs, z)
    #
    # def rho0_c_NFW(self, c, z_eval_rho=0., N=200):
    #     """
    #     computes density normalization as a function of concentration parameter
    #
    #     :param c: concentration
    #     :param z_eval_rho: redshift at which to evaluate the critical density
    #     :param N: the density contrast used to define the halo mass
    #     :return: density normalization in h^2/Mpc^3 (comoving)
    #     """
    #
    #     rho_crit = self.cosmo.rho_crit(z_eval_rho) / self.cosmo.h ** 2
    #     return N / 3 * rho_crit * c ** 3 / (numpy.log(1 + c) - c / (1 + c))
    #
    # def rN_M_nfw_comoving(self, M, N, z):
    #     """
    #     computes the radius R_N of a halo of mass M in comoving distances
    #     :param M: halo mass in M_sun/h
    #     :type M: float or numpy array
    #     :return: radius R_200 in comoving Mpc/h
    #     """
    #
    #     rho_crit = self.cosmo.rho_crit(z) / self.cosmo.h ** 2
    #     return (3 * M / (4 * numpy.pi * rho_crit * N)) ** (1. / 3.)
    #
    # def nfwParam_physical_Mpc(self, M, c, z, N=200):
    #
    #     """
    #
    #     :param M: halo mass in units M_sun (no little h)
    #     :param c: concentration parameter
    #     :param z: redshift
    #     :return: physical rho_s, rs for the NFW profile in comoving units
    #     Mass definition critical density of Universe with respect to critical density at redshift z
    #     Also specified in colossus as 200c
    #     """
    #
    #     h = self.cosmo.h
    #     r200 = self.rN_M_nfw_comoving(M * h, N, z) / h  # comoving virial radius
    #     rhos = self.rho0_c_NFW(c, z, N) * h ** 2  # density in M_sun/Mpc**3
    #     rs = r200 / c
    #     return rhos, rs, r200

    ##################################################################################
    """Routines relevant for other lensing by other mass profiles"""
    ##################################################################################

    def point_mass_factor_z(self, z):

        """
        Returns the cosmology-dependent factor to evaluate the Einstein radius of a point mass of mass M:

        :param z: redshift
        :return: The factor that when multiplied by sqrt(mass) gives the Einstein radius of a point mass

        R_ein = sqrt(M) * point_mass_factor_z(z)

        """
        factor = 4 * self.cosmo.G * self.cosmo.c ** -2
        dds = self.cosmo.D_A(z, self.z_source)
        dd = self.cosmo.D_A_z(z)
        ds = self.D_s
        factor *= dds / dd / ds
        return factor ** 0.5 / self.cosmo.arcsec

    def halo_dynamical_time(self, m_host, z, c_host):
        """
        This routine computes the dynamical timescale for a halo of mass M defined as
        t = 0.5427 / sqrt(G*rho)
        where G is the gravitational constant and rho is the average density
        :param m_host: host mass in M_sun
        :param z: host redshift
        :param c_host: host halo concentration
        :return: the dynamical timescale in Gyr
        """

        _, _, rvir = self.NFW_params_physical(m_host, c_host, z)
        volume = (4/3)*numpy.pi*rvir**3
        rho_average = m_host / volume
        g = 4.3e-6
        return 0.5427 / numpy.sqrt(g*rho_average)

    ##################################################################################
    """ACCRETION REDSHIFT PDF FROM GALACTICUS"""
    ##################################################################################

    @property
    def _subhalo_accretion_pdfs(self):

        if self._computed_zacc_pdf is False:

            self._computed_zacc_pdf = True
            self._mlist, self._dzvals, self._cdfs = self._Msub_cdfs(self.z_lens)

        return self._mlist, self._dzvals, self._cdfs

    def z_accreted_from_zlens(self, msub, zlens):

        mlist, dzvals, cdfs = self._subhalo_accretion_pdfs

        idx = self._mass_index(msub, mlist)

        z_accreted = zlens + self._sample_cdf_single(cdfs[idx])

        return z_accreted

    def _cdf_numerical(self, m, z_lens, delta_z_values):

        c_d_f = []

        prob = 0
        for zi in delta_z_values:
            prob += self._P_fit_diff_M_sub(z_lens + zi, z_lens, m)
            c_d_f.append(prob)
        return numpy.array(c_d_f) / c_d_f[-1]

    def _Msub_cdfs(self, z_lens):

        M_sub_exp = numpy.arange(6.0, 10.2, 0.2)
        M_sub_list = 10 ** M_sub_exp
        delta_z = numpy.linspace(0., 6, 8000)
        funcs = []

        for mi in M_sub_list:
            # cdfi = P_fit_diff_M_sub_cumulative(z_lens+delta_z, z_lens, mi)
            cdfi = self._cdf_numerical(mi, z_lens, delta_z)

            funcs.append(interp1d(cdfi, delta_z))

        return M_sub_list, delta_z, funcs

    def z_decay_mass_dependence(self, M_sub):
        # Mass dependence of z_decay.
        a = 3.21509397
        b = 1.04659814e-03

        return a - b * numpy.log(M_sub / 1.0e6) ** 3

    def z_decay_exp_mass_dependence(self, M_sub):
        # Mass dependence of z_decay_exp.

        a = 0.30335749
        b = 3.2777e-4

        return a - b * numpy.log(M_sub / 1.0e6) ** 3

    def _P_fit_diff_M_sub(self, z, z_lens, M_sub):
        # Given the redhsift of the lens, z_lens, and the subhalo mass, M_sub, return the
        # posibility that the subhlao has an accretion redhisft of z.

        z_decay = self.z_decay_mass_dependence(M_sub)
        z_decay_exp = self.z_decay_exp_mass_dependence(M_sub)

        normalization = 2.0 / numpy.sqrt(2.0 * numpy.pi) / z_decay \
                        / numpy.exp(0.5 * z_decay ** 2 * z_decay_exp ** 2) \
                        / erfc(z_decay * z_decay_exp / numpy.sqrt(2.0))
        return normalization * numpy.exp(-0.5 * ((z - z_lens) / z_decay) ** 2) \
               * numpy.exp(-z_decay_exp * (z - z_lens))

    def _sample_cdf_single(self, cdf_interp):

        u = numpy.random.uniform(0, 1)

        try:
            output = float(cdf_interp(u))
            if numpy.isnan(output):
                output = 0

        except:
            output = 0

        return output

    def _mass_index(self, subhalo_mass, mass_array):

        idx = numpy.argmin(numpy.absolute(subhalo_mass - mass_array))
        return idx



import numpy as np
from pyHalo.Halos.HaloModels.powerlaw import PowerLawSubhalo, PowerLawFieldHalo
from pyHalo.Halos.HaloModels.generalized_nfw import GeneralNFWSubhalo, GeneralNFWFieldHalo
from pyHalo.single_realization import Realization
from pyHalo.Halos.HaloModels.gaussian import Gaussian
from pyHalo.Rendering.correlated_structure import CorrelatedStructure
from pyHalo.Rendering.MassFunctions.delta_function import DeltaFunction
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.utilities import generate_lens_plane_redshifts
from pyHalo.Rendering.SpatialDistributions.uniform import Uniform
from copy import deepcopy


class RealizationExtensions(object):

    """
    This class supports operations that modify individual instances of the class Realization
    (see pyHalo.single_realization).
    """

    def __init__(self, realization):

        """

        :param realization: an instance of Realization
        """

        self._realization = realization

    def core_collapse_by_mass(self, mass_ranges_subhalos, mass_ranges_field_halos,
                              probabilities_subhalos, probabilities_field_halos,
                              kwargs_sub=None, kwargs_field=None):

        """
        This routine transforms some fraction of subhalos and field halos into core collapsed profiles
        in the specified mass ranges

        :param mass_ranges_subhalos: a list of lists specifying the log10 halo mass ranges for subhalos
        e.g. mass_ranges_subhalos = [[6, 8], [8, 10]]
        :param mass_ranges_field_halos: a list of lists specifying the halo mass ranges for field halos
        e.g. mass_ranges_subhalos = [[6, 8], [8, 10]]
        :param probabilities_subhalos: a list of lists specifying the fraction of subhalos in each mass
        range that core collapse
        e.g. probabilities_subhalos = [0.5, 1.] makes half of subhalos with mass 10^6 - 10^8 collapse, and
        100% of subhalos with mass between 10^8 and 10^10 collapse

        If the entries in probabilities subhalos is a function, i.e. you pass in [function_1, function_2]
        instead of [0.5, 1.0], they functions will be evaluated at z_eval, i.e. the fraction of collpsed
        halos becomes [function_1(z_eval), function_2(z_eval)]

        :param probabilities_field_halos: same as probabilities subhalos, but for the population of field halos
        :param kwargs_sub: list of keyword arguments for each function passed in mass_ranges_subhalos, if the entries in
        mass_ranges_subhalos are callable functions
        :param kwargs_field: list of keyword arguments for each function passed in mass_ranges_field_halos, if the entries in
        mass_ranges_field_halos are callable functions
        :return: indexes of core collapsed halos
        """

        assert len(mass_ranges_subhalos) == len(probabilities_subhalos)
        assert len(mass_ranges_field_halos) == len(probabilities_field_halos)

        indexes = []

        for i_halo, halo in enumerate(self._realization.halos):

            u = np.random.rand()

            if halo.is_subhalo:
                for i, mrange in enumerate(mass_ranges_subhalos):
                    if halo.mass >= 10**mrange[0] and halo.mass < 10**mrange[1]:
                        if callable(probabilities_subhalos[i]):
                            prob = probabilities_subhalos[i](halo.z, **kwargs_sub[i])
                        else:
                            prob = probabilities_subhalos[i]
                        if u <= prob:
                            indexes.append(i_halo)
                        break
            else:
                for i, mrange in enumerate(mass_ranges_field_halos):
                    if halo.mass >= 10**mrange[0] and halo.mass < 10**mrange[1]:
                        if callable(probabilities_field_halos[i]):
                            prob = probabilities_field_halos[i](halo.z, **kwargs_field[i])
                        else:
                            prob = probabilities_field_halos[i]
                        if u <= prob:
                            indexes.append(i_halo)
                        break
        return indexes

    def add_core_collapsed_halos(self, indexes, halo_profile='SPL_CORE',**kwargs_halo):

        """
        This function turns NFW halos in a realization into profiles modeled as PseudoJaffe profiles
        with 1/r^2 central density profiles with the same total mass as the original NFW
        profile.

        :param indexes: the indexes of halos in the realization to transform into powerlaw or generalized nfw profiles
        :param halo_profile: specifies whether to transform to powerlaw (SPL_CORE) or generalized_nfw profile (GNFW)
        :param kwargs_halo: the keyword arguments for the collapsed halo profile
        :return: A new instance of Realization where the halos indexed by indexes
        in the original realization have their mass definitions changed to PsuedoJaffe
        """

        halos = self._realization.halos
        new_halos = []

        if halo_profile == 'GNFW':
            subhalo_model = GeneralNFWSubhalo
            fieldhalo_model = GeneralNFWFieldHalo
        elif halo_profile == 'SPL_CORE':
            subhalo_model = PowerLawSubhalo
            fieldhalo_model = PowerLawFieldHalo
        else:
            raise Exception('only halo profile models GNFW and SPL_CORE '
                            'implemented for collapsed objects')

        for i, halo in enumerate(halos):
            if i in indexes:
                halo._args.update(kwargs_halo)
                if halo.is_subhalo:
                    new_halo = subhalo_model(halo.mass, halo.x, halo.y, halo.r3d,
                                                 halo.z, True, halo.lens_cosmo, halo._args,
                                             halo._truncation_class,
                                             halo._concentration_class,
                                             halo.unique_tag)
                else:
                    new_halo = fieldhalo_model(halo.mass, halo.x, halo.y, halo.r3d,
                                                 halo.z, False, halo.lens_cosmo, halo._args,
                                               halo._truncation_class,
                                               halo._concentration_class,
                                               halo.unique_tag)
                new_halos.append(new_halo)
            else:
                new_halos.append(halo)

        return Realization.from_halos(new_halos, self._realization.lens_cosmo, self._realization.kwargs_halo_model,
                                      self._realization.apply_mass_sheet_correction,
                                      self._realization.rendering_classes,
                                      self._realization._rendering_center_x, self._realization._rendering_center_y,
                                      self._realization.geometry)

    def add_ULDM_fluctuations(self, de_Broglie_wavelength, fluctuation_amplitude,
                              fluctuation_size, fluctuation_size_variance, n_cut, n_fluc_scale=1., shape='ring',
                              args={'rmin':0.9,'rmax':1.1}, rescale_fluc_amp=True):

        """
        This function adds gaussian fluctuations of the given de Broglie wavelength to a realization.

        :param de_Broglie_wavelength: de Broglie wavelength of the DM particle in kpc
        :param fluctuation_amplitude: Standard deviation of amplitude distribution in convergence units
        :param fluctuation_size: half the physical size of an indiviudal blob [kpc]
        (Blobs are modeled as Gaussians, so their diameter is ~2 * fluctuation_size)
        :param fluctuation_size_variance: scales the variance of the distribution of sizes of fluctuations,
        relative to lambda_dB
        :param shape: keyword argument for fluctuation geometry, can place fluctuations in a

            'ring', 'ellipse, or 'aperture' (centered about lensing images)

        :param args: properties of the given shape, must match

            'ring' : {'rmin': , 'rmax': } (radii within which to place fluctuations, rmin < rmax)
            'ellipse' : {'amin': , 'amax': , 'bmin': , 'bmax': , 'angle':} (major and minor axes within which to place fluctuations, amin < amax, bmin < bmax)
            'aperture' : {'x_images': , 'y_images':, 'aperture'} (list of x and y image coordinates and aperture radius)

            Note that for 'ellipse' the 'angle' parameter is the angle in radians at which to orient the ellipse relative to the positive x-axis.

        :param num_cut: integer number of fluctuations above which to start cancelling fluctuations
        :param rescale_fluc_amp: Boolean specifying whether re-scale fluctuation amplitudes by sqrt(n_cut / n), where n
        is the total number of fluctuations in the given area and n_cut (defined below) is the maximum number to generate
        """

        if (shape != 'ring') and (shape != 'ellipse') and (shape != 'aperture'): # check shape keyword
            raise Exception('shape must be ring or ellipse or aperture!')

        # get number of fluctuations
        n_flucs = _get_number_flucs(self._realization, de_Broglie_wavelength,
                                    fluctuation_size/de_Broglie_wavelength, n_fluc_scale, shape, args)

        # if zero fluctuations, return original realization

        if shape == 'aperture':
            for i in range(0, len(n_flucs)):
                if n_flucs[i] != 0:
                    break
            else:
                return self._realization

        else:
            if n_flucs == 0:
                return self._realization

        if rescale_fluc_amp:
            if shape == 'aperture' and np.mean(n_flucs) > n_cut:
                fluctuation_amplitude /= np.sqrt(np.mean(n_flucs) / n_cut)
                n_flucs = np.array([int(n_cut)] * 4)
            elif shape != 'aperture' and n_flucs > n_cut:
                fluctuation_amplitude /= np.sqrt(n_flucs / n_cut)
                n_flucs = int(n_cut)

        # create fluctuations
        fluctuations = _get_fluctuation_halos(self._realization,
                                              fluctuation_amplitude,
                                              fluctuation_size,
                                              fluctuation_size_variance,
                                              shape,
                                              n_flucs,
                                              args)

        # realization args
        lens_cosmo = self._realization.lens_cosmo
        msheet_correction = self._realization._mass_sheet_correction
        rendering_classes = self._realization.rendering_classes
        rendering_center_x, rendering_center_y = self._realization.rendering_center
        kwargs_halo_model = None
        # realization containing only fluctuations
        fluc_realization = Realization.from_halos(fluctuations, lens_cosmo, kwargs_halo_model,
                                      msheet_correction, rendering_classes,
                                      rendering_center_x, rendering_center_y,
                                      self._realization.geometry)

        # join realization to dark substructure realization
        return self._realization.join(fluc_realization)

    def add_correlated_structure(self, mass_function_model,
                                 kwargs_mass_function,
                                 mass_definition,
                                   x_image_interp_list,
                                   y_image_interp_list,
                                   arcsec_per_pixel, r_max,
                                 rescale_normalizations=True):

        """
        Adds structure along the line of sight with a spatial distribution that tracks the dark matter density at each
        lens plane

        """

        realization_copy = deepcopy(self._realization)
        lens_plane_redshifts, delta_z_list = generate_lens_plane_redshifts(self._realization.lens_cosmo.z_lens,
                                                                           self._realization.lens_cosmo.z_source)
        correlated_structure = CorrelatedStructure(mass_function_model, kwargs_mass_function,
                 self._realization.geometry, self._realization.lens_cosmo, lens_plane_redshifts,
                                                   delta_z_list, self._realization)

        for image_index, (x_image_interp, y_image_interp) in enumerate(zip(x_image_interp_list, y_image_interp_list)):
            masses, x, y, r3d, redshifts, subhalo_flag, rescale_indicies, rescale_factors = correlated_structure.render(
                r_max, x_image_interp, y_image_interp, arcsec_per_pixel)

            if rescale_normalizations:
                for i, index in enumerate(rescale_indicies):
                    realization_copy.halos[index].rescale_normalization(rescale_factors[i])

            mdefs = [mass_definition] * len(masses)
            kwargs_halo_model = {'truncation_model': None, 'concentration_model': None, 'kwargs_density_profile': {}}
            if image_index==0:
                realization_pbh = Realization(masses, x, y, r3d, mdefs, redshifts, subhalo_flag,
                                           self._realization.lens_cosmo, kwargs_halo_model=kwargs_halo_model)
            elif image_index>0:
                realization_pbh = realization_pbh.join(Realization(masses, x, y, r3d, mdefs, redshifts, subhalo_flag,
                                           self._realization.lens_cosmo, kwargs_halo_model=kwargs_halo_model))

        new_realization = realization_copy.join(realization_pbh)

        return new_realization

    def add_primordial_black_holes(self, pbh_mass_fraction, kwargs_pbh_mass_function, mass_fraction_in_halos,
                                   x_image_interp_list, y_image_interp_list, r_max_arcsec, arcsec_per_pixel=0.005,
                                   rescale_normalizations=True):

        """
        This routine renders populations of primordial black holes modeled as point masses along the line of sight.
        The population of objects includes a smoothly distributed component, and a component that is clustered according
        to the population of halos generated in the instance of Realization used to instantiate the class.

        :param pbh_mass_fraction: the mass fraction of dark matter contained in primordial black holes
        :param kwargs_pbh_mass_function: keyword arguments for the PBH mass function
        :param mass_fraction_in_halos: the fraction of dark matter mass contained in halos in the mass range
        used to generate the instance of realization used to instantiate the class
        :param x_image_interp_list: a list of interp1d functions that return the angular x coordinate of a light ray
        given a comoving distance
        :param y_image_interp_list: a list of interp1d functions that return the angular y coordinate of a light ray
        given a comoving distance
        :param r_max_arcsec: the radius of the rendering region in arcsec, here an array corresponding to the coordinates used for x_interp_list
        :param arcsec_per_pixel: the resolution of the grid used to compute the population of PBH whose spatial
        distribution tracks the dark matter density along the LOS specific by the instance of Realization used to
        instantiate the class
        :param rescale_normalizations: bool; whether or not to rescale the density profile of halos to account for the
        mass added in correlated structure
        :return: a new instance of Realization that contains primordial black holes modeled as point masses
        """
        mass_definition = 'PT_MASS'
        plane_redshifts = self._realization.unique_redshifts
        delta_z = []
        for i, zi in enumerate(plane_redshifts[0:-1]):
            delta_z.append(plane_redshifts[i + 1] - plane_redshifts[i])
        delta_z.append(self._realization.lens_cosmo.z_source - plane_redshifts[-1])

        mass_fraction_smooth = (1 - mass_fraction_in_halos) * pbh_mass_fraction
        mass_fraction_clumpy = pbh_mass_fraction * mass_fraction_in_halos

        masses = np.array([])
        xcoords = np.array([])
        ycoords = np.array([])
        redshifts = np.array([])
        for x_image_interp, y_image_interp, r_max in zip(x_image_interp_list, y_image_interp_list, r_max_arcsec):
            geometry = Geometry(self._realization.lens_cosmo.cosmo,
                                        self._realization.lens_cosmo.z_lens,
                                        self._realization.lens_cosmo.z_source,
                                        2 * r_max,
                                        'DOUBLE_CONE')
            for zi, delta_zi in zip(plane_redshifts, delta_z):

                d = self._realization.lens_cosmo.cosmo.D_C_transverse(zi)
                angle_x, angle_y = x_image_interp(d), y_image_interp(d)
                rendering_radius = r_max * geometry.rendering_scale(zi)
                spatial_distribution_model_smooth = Uniform(rendering_radius, geometry)

                if kwargs_pbh_mass_function['mass_function_type'] == 'DELTA':
                    rho_smooth = mass_fraction_smooth * self._realization.lens_cosmo.cosmo.rho_dark_matter_crit
                    volume = geometry.volume_element_comoving(zi, delta_zi)
                    mass_function_smooth = DeltaFunction(10 ** kwargs_pbh_mass_function['logM'],
                                                         volume, rho_smooth, draw_poisson=True)
                else:
                    raise Exception('no mass function type for PBH currently implemented besides DELTA')

                m_smooth = mass_function_smooth.draw()
                if len(m_smooth) > 0:
                    kpc_per_asec = geometry.kpc_per_arcsec(zi)
                    x_kpc, y_kpc = spatial_distribution_model_smooth.draw(len(m_smooth), zi,
                                                                          center_x=angle_x, center_y=angle_y)
                    x_arcsec, y_arcsec = x_kpc / kpc_per_asec, y_kpc / kpc_per_asec
                    masses = np.append(masses, m_smooth)
                    xcoords = np.append(xcoords, x_arcsec)
                    ycoords = np.append(ycoords, y_arcsec)
                    redshifts = np.append(redshifts, np.array([zi] * len(m_smooth)))

        mdefs = [mass_definition] * len(masses)
        r3d = np.array([None] * len(masses))
        subhalo_flag = [False] * len(masses)
        kwargs_halo_model = {'truncation_model': None, 'concentration_model': None, 'kwargs_density_profile': {}}
        realization_smooth = Realization(masses, xcoords, ycoords, r3d, mdefs, redshifts, subhalo_flag,
                                         self._realization.lens_cosmo, kwargs_halo_model=kwargs_halo_model,
                                         mass_sheet_correction=False)

        kwargs_pbh_mass_function['mass_fraction'] = mass_fraction_clumpy

        for ii, (x_image_interp, y_image_interp, r_max) in enumerate(zip(x_image_interp_list, y_image_interp_list, r_max_arcsec)):
            realization_with_clustering_temp = self.add_correlated_structure(DeltaFunction,
                                 kwargs_pbh_mass_function,
                                 mass_definition,
                                   [x_image_interp],
                                   [y_image_interp],
                                   arcsec_per_pixel, r_max,
                                 rescale_normalizations)
            if ii == 0:
                realization_with_clustering = realization_with_clustering_temp
                continue # first time through loop
            realization_with_clustering = realization_with_clustering.join(realization_with_clustering_temp)

        return realization_with_clustering.join(realization_smooth)

def _get_number_flucs(realization, de_Broglie_wavelength, fluctuation_size_scale, n_fluc_scale, shape, args):
    """
    This function returns the number of fluctuations to place in the realization.

    :param realization: the realization to which to add the fluctuations
    :param de_Broglie_wavelength: de Broglie wavelength of the DM particle in kpc
    :param fluctuation_size_scale: sets the size of the fluctuatiions relative to the de Broglie wavelength
    :param n_fluc_scale: rescales the total number of fluctuations
    :param shape: keyword argument for fluctuation geometry, see 'add_ULDM_fluctuations'
    :param args: properties of the given shape, see 'add_ULDM_fluctuations'
    """

    D_d = realization.lens_cosmo.cosmo.D_A_z(realization._zlens)
    arcsec = realization.lens_cosmo.cosmo.arcsec
    fluc_area=np.pi*(de_Broglie_wavelength * fluctuation_size_scale)**2 #de Broglie area
    to_kpc = D_d * arcsec * 1e3 # convert arcsec to kpc

    if shape=='ring': # fluctuations in a circular slice (for visualization purposes)

        rmin_kpc,rmax_kpc = args['rmin'] * to_kpc, args['rmax'] * to_kpc #args in kpc
        area_ring = np.pi*(rmax_kpc**2-rmin_kpc**2) # volume of ring
        n_flucs_expected=n_fluc_scale*area_ring/fluc_area # number of fluctuations in ring
        n_flucs = np.random.poisson(n_flucs_expected)

    if shape=='ellipse': # fluctuations in a elliptical slice (for visualization purposes)

        amin_kpc,bmin_kpc,amax_kpc,bmax_kpc=args['amin'] * to_kpc, args['bmin'] * to_kpc, args['amax'] * to_kpc, args['bmax'] * to_kpc #args in kpc
        area_ellipse=np.pi*(amax_kpc*bmax_kpc - amin_kpc*bmin_kpc) # volume of ellipse
        n_flucs_expected=n_fluc_scale*area_ellipse/fluc_area # number of fluctuations in ellipse
        n_flucs = np.random.poisson(n_flucs_expected)

    if shape=='aperture': # fluctuations around lensing images (for computation)

        n_images=len(args['x_images']) #number of lensed images
        r_kpc = args['aperture'] * to_kpc #aperture in kpc
        area_aperture = np.pi*r_kpc**2 # aperture area
        n_flucs_expected = n_fluc_scale*area_aperture/fluc_area #number of expected fluctuations per aperture
        n_flucs = np.random.poisson(n_flucs_expected,n_images) #draw number of fluctuations from poisson distribution for each image
        n_flucs = n_flucs[n_flucs!=0] #get rid of aperture if zero fluctuations within it

    return n_flucs

def _get_fluctuation_halos(realization, fluctuation_amplitude, fluctuation_size, fluctuation_size_variance, shape, n_flucs, args):
    """
    This function creates 'n_flucs' Gaussian fluctuations and places them according to 'shape'.

    :param realization: the realization to which to add the fluctuations
    :param fluctuation_amplitude: Standard deviation of amplitude distribution in convergence units
    :param fluctuation_size: half the physical size of a fluctuation (individual blobs are modeled as Gaussians, with
    a variance fluctuation_size)
    :param fluctuation_size_variance: scales the variance of the distribution of sizes of fluctuations, relative to lambda_dB
    :param rho_bar: typical convergence of a fluctuation at its peak
    :param shape: keyword argument for fluctuation geometry, see 'add_ULDM_fluctuations'
    :param n_flucs: Number of fluctuations to make
    :param args: properties of the given shape, see 'add_ULDM_fluctuations'
    """

    kpc_per_arcsec = realization.lens_cosmo.cosmo.kpc_proper_per_asec(realization._zlens)
    fluc_var_angle = fluctuation_size / kpc_per_arcsec # gaussian variance in arcsec
    fluctuation_size_variance_angle = fluctuation_size_variance / kpc_per_arcsec

    if shape != 'aperture':

        sigs = np.abs(np.random.normal(fluc_var_angle,fluctuation_size_variance_angle,n_flucs)) #random widths
        kappa0 = np.random.normal(0, fluctuation_amplitude, n_flucs)
        # kappa0 = amp / (2 * np.pi * sigma ** 2)
        amps = kappa0 * 2 * np.pi * sigs ** 2

    if shape=='ring':
        angles = np.random.uniform(0,2*np.pi,n_flucs)  # random angles
        radii = args['rmin'] + np.sqrt(np.random.uniform(0,1,n_flucs))*(args['rmax']-args['rmin']) #random radii
        xs = radii*np.cos(angles) #random x positions
        ys = radii*np.sin(angles) #random y positions

    if shape=='ellipse':
        angles = np.random.uniform(0,2*np.pi,n_flucs)  # random angles
        aa = np.sqrt(np.random.uniform(0,1,n_flucs))*(args['amax']-args['amin']) + args['amin'] #random axis 1
        bb = np.sqrt(np.random.uniform(0,1,n_flucs))*(args['bmax']-args['bmin']) + args['bmin'] #random axis 1
        xs = aa*np.cos(angles)*np.cos(args['angle'])-bb*np.sin(angles)*np.sin(args['angle']) #random x positions
        ys = aa*np.cos(angles)*np.sin(args['angle'])+bb*np.sin(angles)*np.cos(args['angle']) #random y positions

    if shape == 'aperture':

        amps, sigs, xs, ys = np.array([]), np.array([]), np.array([]), np.array([])

        for i in range(0, len(n_flucs)): #loop through each image

            sigs_i = np.random.normal(fluc_var_angle,fluctuation_size_variance_angle,n_flucs[i])
            sigs_i = np.absolute(sigs_i)

            kappa0 = np.random.normal(0, fluctuation_amplitude, n_flucs[i])
            amps_i = kappa0 * 2*np.pi*sigs_i**2

            angles_i = np.random.uniform(0, 2*np.pi, n_flucs[i])  # random angles
            r = np.random.uniform(0, args['aperture'] ** 2, int(n_flucs[i]))
            xs_i = r ** 0.5 * np.sin(angles_i) + args['x_images'][i]
            ys_i = r ** 0.5 * np.cos(angles_i) + args['y_images'][i]
            amps, sigs, xs, ys= np.append(amps, amps_i), np.append(sigs, sigs_i), np.append(xs, xs_i), np.append(ys, ys_i)

    args_fluc=[{'amp': amps[i], 'sigma': sigs[i], 'center_x': xs[i], 'center_y': ys[i]} for i in range(len(amps))]
    # kappa(r) = kappa * exp(-0.5 * r^2/sigma^2)
    #sigma_crit = realization.lens_cosmo.sigmacrit # in units M_sun / arcsec^2
    masses = np.absolute(amps)
    #masses = [10 ** 8.0] * len(xs)
    fluctuations = [Gaussian(masses[i], xs[i], ys[i], None, realization.lens_cosmo.z_lens,
                             True, realization.lens_cosmo, args_fluc[i],
                             truncation_class=None, concentration_class=None,
                             unique_tag=np.random.rand()) for i in range(len(amps))]

    return fluctuations



from pyHalo.single_realization import Realization, SingleHalo, realization_at_z
from lenstronomy.LensModel.lens_model import LensModel
import numpy as np
import numpy.testing as npt
from scipy.interpolate import interp1d
from copy import deepcopy
import pytest
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce, ConcentrationPeakHeight
from pyHalo.Halos.tidal_truncation import TruncationRN
from pyHalo.preset_models import CDM
from pyHalo.Halos.lens_cosmo import LensCosmo
import matplotlib.pyplot as plt


class TestSingleRealization(object):

    def setup_method(self):
        z_lens = 0.5
        z_source = 1.5
        self._logmlow = 6.0
        self._logmhigh = 9.0
        self.realization = CDM(z_lens, z_source, log_mlow=self._logmlow, log_mhigh=self._logmhigh, sigma_sub=0.05,
                                   cone_opening_angle_arcsec=12.0)
        self.rendering_classes = self.realization.rendering_classes


    def test_mass_sheet_correction(self):

        kwargs_mass_sheets, profile_list, z_sheets = self.realization._mass_sheet_correction(self.rendering_classes,
                                                                                             subtract_exact_sheets=False,
                                                                                             kappa_scale=1.0,
                                                                                             log_mlow_sheets=self._logmlow,
                                                                                             log_mhigh_sheets=self._logmhigh)

        npt.assert_equal(len(kwargs_mass_sheets), len(profile_list))
        npt.assert_equal(len(kwargs_mass_sheets), len(z_sheets))

        profile_list, lens_redshift_list, kwargs_halos, _ = self.realization.lensing_quantities()
        lens_model = LensModel(profile_list, lens_redshift_list=list(lens_redshift_list),
                               z_source=self.realization.lens_cosmo.z_source,
                               multi_plane=True, cosmo=self.realization.lens_cosmo.cosmo.astropy)
        _r = np.linspace(-2.5, 2.5, 100)
        xx, yy = np.meshgrid(_r, _r)
        kappa = lens_model.kappa(xx.ravel(), yy.ravel(), kwargs_halos)
        mean_kappa = np.mean(kappa)
        npt.assert_array_less(abs(mean_kappa), 0.06)

        kwargs_mass_sheets, profile_list, z_sheets = self.realization._mass_sheet_correction(self.rendering_classes,
                                                                                             subtract_exact_sheets=True,
                                                                                             kappa_scale=1.0,
                                                                                             log_mlow_sheets=self._logmlow,
                                                                                             log_mhigh_sheets=self._logmhigh)

        npt.assert_equal(len(kwargs_mass_sheets), len(profile_list))
        npt.assert_equal(len(kwargs_mass_sheets), len(z_sheets))
        profile_list, lens_redshift_list, kwargs_halos, _ = self.realization.lensing_quantities()
        lens_model = LensModel(profile_list, lens_redshift_list=list(lens_redshift_list),
                               z_source=self.realization.lens_cosmo.z_source,
                               multi_plane=True, cosmo=self.realization.lens_cosmo.cosmo.astropy)
        _r = np.linspace(-2.5, 2.5, 100)
        xx, yy = np.meshgrid(_r, _r)
        kappa = lens_model.kappa(xx.ravel(), yy.ravel(), kwargs_halos)
        mean_kappa = np.mean(kappa)
        npt.assert_array_less(abs(mean_kappa), 0.04)

    def test_build_from_halos(self):

        realization_fromhalos = Realization.from_halos(self.realization.halos, self.realization.lens_cosmo,
                                   self.realization.kwargs_halo_model,
                                                       self.realization.apply_mass_sheet_correction,
                                                       self.realization.rendering_classes)

        for halo_1, halo_2 in zip(realization_fromhalos.halos, self.realization.halos):
            npt.assert_equal(halo_1.x, halo_2.x)
            npt.assert_equal(halo_1.y, halo_2.y)
            npt.assert_equal(halo_1.mass, halo_2.mass)
            npt.assert_equal(halo_1.r3d, halo_2.r3d)
            npt.assert_string_equal(halo_1.mdef, halo_2.mdef)
            npt.assert_equal(halo_1.z, halo_2.z)
            npt.assert_equal(halo_1.is_subhalo, halo_2.is_subhalo)
            npt.assert_equal(halo_1.unique_tag, halo_2.unique_tag)

        npt.assert_equal(True, realization_fromhalos == self.realization)
        npt.assert_equal(realization_fromhalos.apply_mass_sheet_correction,
                         self.realization.apply_mass_sheet_correction)

    def test_join(self):

        z_lens = 0.5
        z_source = 1.5
        realization_2 = CDM(z_lens, z_source, log_mlow=self._logmlow, log_mhigh=self._logmhigh, sigma_sub=0.05,
                                 cone_opening_angle_arcsec=12.0, LOS_normalization=0.0)
        new_realization = self.realization.join(realization_2, join_rendering_classes=True)
        npt.assert_equal(len(new_realization.rendering_classes), 6)

        length = len(realization_2.halos) + len(self.realization.halos)
        npt.assert_equal(len(new_realization.halos), length)
        # now change two of the tags to be the same as in realization_cdm, make sure they're gone

        idx2_1 = 0
        idx2_2 = 1
        idx1 = 0
        idx2 = 1
        self.realization.halos[idx1].unique_tag = realization_2.halos[idx2_1].unique_tag
        self.realization.halos[idx2].unique_tag = realization_2.halos[idx2_2].unique_tag
        new_realization = self.realization.join(realization_2)
        npt.assert_equal(len(new_realization.halos), length-2)

        original_x_list = np.append(self.realization.x, realization_2.x)
        original_m_list = np.append(self.realization.masses, realization_2.masses)

        for (xi, mi) in zip(original_x_list, original_m_list):
            if xi not in new_realization.x:
                npt.assert_equal(True, mi not in new_realization.masses)
            if mi not in new_realization.masses:
                npt.assert_equal(True, xi not in new_realization.x)

    def test_shift_background_to_source(self):

        dmax = self.realization.geometry.cosmo.D_C_transverse(self.realization.lens_cosmo.z_source)
        z = np.linspace(0, dmax, 100)
        ray_interp_x, ray_interp_y = interp1d(z, np.ones_like(z)), interp1d(z, -np.ones_like(z))

        centerx_interp, centery_interp = self.realization.rendering_center
        npt.assert_almost_equal(centerx_interp(1000), 0.)
        npt.assert_almost_equal(centery_interp(1000), 0.)

        model_name, _, kwargs_lens_init, _ = self.realization.lensing_quantities()
        for name, kw in zip(model_name, kwargs_lens_init):
            if name == 'CONVERGENCE':
                npt.assert_almost_equal(kw['ra_0'], 0.)
                npt.assert_almost_equal(kw['dec_0'], 0.)

        test_realization = deepcopy(self.realization)
        realization_shifted = test_realization.shift_background_to_source(ray_interp_x, ray_interp_y)

        for halo, halo_0 in zip(realization_shifted.halos, self.realization.halos):

            npt.assert_equal(halo.x, halo_0.x + 1)
            npt.assert_equal(halo.y, halo_0.y - 1)

        model_names, redshifts_lens, kwargs_lens_init, _ = self.realization.lensing_quantities()
        centerx, centerey = self.realization.rendering_center
        for kw, zi, name in zip(kwargs_lens_init, redshifts_lens, model_names):
            if name == 'CONVERGENCE':
                di = self.realization.lens_cosmo.cosmo.D_C_z(zi)
                npt.assert_almost_equal(kw['ra_0'], centerx(di))
                npt.assert_almost_equal(kw['dec_0'], centerey(di))

        # make sure you can only shift realizations once
        realization_shifted = realization_shifted.shift_background_to_source(ray_interp_x, ray_interp_y)
        for halo, halo_0 in zip(realization_shifted.halos, self.realization.halos):
            npt.assert_equal(halo.x, halo_0.x + 1)
            npt.assert_equal(halo.y, halo_0.y - 1)

    def test_filter(self):

        masses = [10 ** 9.04, 10 ** 8.2, 10 ** 7.345, 10 ** 7, 10 ** 6.05, 10 ** 7, 10 ** 7]
        x = [0.5, 0.1, -0.9, -1.4, 1.2, 0., -1.]
        y = [0., 0.9, -1.4, -1., -0.4, 1., 0.9]
        redshifts = [0.1, 0.2, 0.3, 0.4, 0.4, 0.94, 0.94]
        subflags = [False, False, False, True, True, False, False]
        r3d = [100] * len(masses)
        mdefs_TNFW = ['TNFW'] * len(masses)

        realization_cdm = Realization(masses, x, y, r3d, mdefs_TNFW, redshifts, subflags,
                                      self.realization.lens_cosmo, halos=None, kwargs_halo_model=self.realization.kwargs_halo_model,
                                      mass_sheet_correction=True, rendering_classes=self.realization.rendering_classes)

        r = [self.realization.geometry.cosmo.D_C_transverse(zi) for zi in realization_cdm.redshifts]

        x_intercepts_1 = [0.5, 0.1, -0.9, -1.4, 1.2, 0., -1.]
        y_intercepts_1 = [0., 0.9, -1.4, -1., -0.4, 1., 0.9]
        x_intercepts_2 = [1000] * len(x_intercepts_1)
        y_intercepts_2 = [1000] * len(y_intercepts_1)
        ray_interp_x1 = interp1d(r, x_intercepts_1)
        ray_interp_y1 = interp1d(r, y_intercepts_1)
        ray_interp_x2 = interp1d(r, x_intercepts_2)
        ray_interp_y2 = interp1d(r, y_intercepts_2)

        aperture_radius_front = 0.3
        aperture_radius_back = 0.3
        log_mass_allowed_in_aperture_front = 6
        log_mass_allowed_in_aperture_back = 6
        log_mass_allowed_global_front = 6
        log_mass_allowed_global_back = 6
        interpolated_x_angle = [ray_interp_x1, ray_interp_x2]
        interpolated_y_angle = [ray_interp_y1, ray_interp_y2]

        realization_filtered = realization_cdm.filter(aperture_radius_front,
                   aperture_radius_back,
                   log_mass_allowed_in_aperture_front,
                   log_mass_allowed_in_aperture_back,
                   log_mass_allowed_global_front,
                   log_mass_allowed_global_back,
                   interpolated_x_angle, interpolated_y_angle)
        npt.assert_equal(realization_filtered == realization_cdm, True)

        x_intercepts_1 = [1000] * len(x_intercepts_1)
        y_intercepts_1 = [1000] * len(x_intercepts_1)
        x_intercepts_2 = [1000] * len(x_intercepts_1)
        y_intercepts_2 = [1000] * len(y_intercepts_1)
        ray_interp_x1 = interp1d(r, x_intercepts_1)
        ray_interp_y1 = interp1d(r, y_intercepts_1)
        ray_interp_x2 = interp1d(r, x_intercepts_2)
        ray_interp_y2 = interp1d(r, y_intercepts_2)

        aperture_radius_front = 0.3
        aperture_radius_back = 0.3
        log_mass_allowed_in_aperture_front = 6
        log_mass_allowed_in_aperture_back = 6
        log_mass_allowed_global_front = 10
        log_mass_allowed_global_back = 10
        interpolated_x_angle = [ray_interp_x1, ray_interp_x2]
        interpolated_y_angle = [ray_interp_y1, ray_interp_y2]

        realization_filtered = realization_cdm.filter(aperture_radius_front,
                                                      aperture_radius_back,
                                                      log_mass_allowed_in_aperture_front,
                                                      log_mass_allowed_in_aperture_back,
                                                      log_mass_allowed_global_front,
                                                      log_mass_allowed_global_back,
                                                      interpolated_x_angle, interpolated_y_angle)
        npt.assert_equal(len(realization_filtered.halos), 0)

        x_intercepts_1 = [1000] * len(x_intercepts_1)
        y_intercepts_1 = [1000] * len(x_intercepts_1)
        x_intercepts_2 = [0.5, 0.1, -0.9, -1.4, 1.2, 0., -1.]
        y_intercepts_2 = [0., 0.9, -1.4, -1., -0.4, 1., 0.9]
        offset = 0.25
        idx_offset = 4
        tag = realization_cdm.halos[idx_offset].unique_tag
        x_intercepts_2[idx_offset] += offset

        offset = 0.35
        idx_offset = 2
        tag2 = realization_cdm.halos[idx_offset].unique_tag
        y_intercepts_2[idx_offset] += offset

        ray_interp_x1 = interp1d(r, x_intercepts_1)
        ray_interp_y1 = interp1d(r, y_intercepts_1)
        ray_interp_x2 = interp1d(r, x_intercepts_2)
        ray_interp_y2 = interp1d(r, y_intercepts_2)

        aperture_radius_front = 0.3
        aperture_radius_back = 0.3
        log_mass_allowed_in_aperture_front = 6.
        log_mass_allowed_in_aperture_back = 6
        log_mass_allowed_global_front = 10
        log_mass_allowed_global_back = 10
        interpolated_x_angle = [ray_interp_x1, ray_interp_x2]
        interpolated_y_angle = [ray_interp_y1, ray_interp_y2]

        realization_filtered = realization_cdm.filter(aperture_radius_front,
                                                      aperture_radius_back,
                                                      log_mass_allowed_in_aperture_front,
                                                      log_mass_allowed_in_aperture_back,
                                                      log_mass_allowed_global_front,
                                                      log_mass_allowed_global_back,
                                                      interpolated_x_angle, interpolated_y_angle)
        new_tags = realization_filtered._tags()
        npt.assert_equal(True, tag in new_tags)
        npt.assert_equal(True, tag2 not in new_tags)

        realization_filtered = realization_cdm.filter(aperture_radius_front,
                                                      aperture_radius_back,
                                                      log_mass_allowed_in_aperture_front,
                                                      log_mass_allowed_in_aperture_back,
                                                      log_mass_allowed_global_front,
                                                      log_mass_allowed_global_back,
                                                      interpolated_x_angle, interpolated_y_angle, aperture_units='MPC')

        npt.assert_equal(len(realization_filtered.halos) > 0, True)

        args = (aperture_radius_front,aperture_radius_back, log_mass_allowed_in_aperture_front,
                        log_mass_allowed_in_aperture_back, log_mass_allowed_global_front,
                        log_mass_allowed_global_back, interpolated_x_angle,
                interpolated_y_angle, None, None, 'nonsense')

        npt.assert_raises(Exception, realization_cdm.filter, *args)

    def test_lensing_quantities(self):

        lens_model_list, redshift_array, kwargs_lens, kwargs_lensmodel = \
            self.realization.lensing_quantities(add_mass_sheet_correction=True)
        npt.assert_equal(True, 'CONVERGENCE' in lens_model_list)

        lens_model_list, redshift_array, kwargs_lens, kwargs_lensmodel = \
            self.realization.lensing_quantities(add_mass_sheet_correction=False)
        npt.assert_equal(True, 'CONVERGENCE' not in lens_model_list)
        npt.assert_equal(len(lens_model_list), len(self.realization.x))

        lens_model_list, redshift_array, kwargs_lens, kwargs_lensmodel = \
            self.realization.lensing_quantities(add_mass_sheet_correction=True)
        npt.assert_equal(True, 'CONVERGENCE' in lens_model_list)

    def test_split_at_z(self):

        real_before, real_after = self.realization.split_at_z(0.4)
        halos_before = real_before.halos
        halos_after = real_after.halos
        npt.assert_equal(len(halos_before), np.sum(np.array(self.realization.redshifts) <= 0.4))
        npt.assert_equal(len(halos_after), np.sum(np.array(self.realization.redshifts) > 0.4))

    def test_single_halo(self):

        kwargs_halo_model = {'concentration_model': None, 'truncation_model': None, 'kwargs_density_profile': {}}
        single_halo = SingleHalo(10**8, 0.5, 1.0, 'PT_MASS', 0.5, 0.5, 1.5, r3d=None, subhalo_flag=False,
                                 kwargs_halo_model=kwargs_halo_model)
        lens_model_list, redshift_array, kwargs_lens, kwargs_lensmodel = single_halo.lensing_quantities()
        npt.assert_equal(len(lens_model_list), 1)
        npt.assert_string_equal(lens_model_list[0], 'POINT_MASS')

    def test_realization_at_z(self):

        realatz, inds = realization_at_z(self.realization, 0.5, 0., 0., 1)
        for halo in realatz.halos:
            npt.assert_equal(True, halo.z == 0.5)
            npt.assert_array_less(np.hypot(halo.x, halo.y), 1.00000000001)

    def test_comoving_coordinates(self):

        x, y, logm, z = self.realization.halo_comoving_coordinates()
        d1 = self.realization.lens_cosmo.cosmo.D_C_z(z[0])
        npt.assert_almost_equal(x[0], self.realization.x[0]*d1)
        npt.assert_almost_equal(y[0], self.realization.y[0] * d1)
        npt.assert_almost_equal(10**logm[0], self.realization.masses[0])

    def test_assign_concentration_models(self):

        m = 10 ** 8
        x = 0.5
        y = 0.1
        mdef = 'TNFW'
        z = 0.5
        zlens = 0.5
        zsource = 1.5
        r3d = None

        lens_cosmo = LensCosmo(zlens, zsource)
        kwargs_halo_model_1 = {'concentration_model': ConcentrationDiemerJoyce(lens_cosmo.cosmo.astropy),
                               'truncation_model': TruncationRN(lens_cosmo), 'kwargs_density_profile': {}}
        subhalo_flag = True
        single_subhalo_diemer_joyce = SingleHalo(10**7, x, y, mdef, z, zlens, zsource, r3d, subhalo_flag,
                                   kwargs_halo_model_1, lens_cosmo=lens_cosmo)
        subhalo_flag = False
        single_fieldhalo_diemer_joyce = SingleHalo(10**8, x, y, mdef, z, zlens, zsource, r3d, subhalo_flag,
                                                 kwargs_halo_model_1, lens_cosmo=lens_cosmo)

        kwargs_halo_model_2 = {'concentration_model': ConcentrationPeakHeight(lens_cosmo.cosmo.astropy, 15.0, -0.1, -0.8),
                             'truncation_model': TruncationRN(lens_cosmo), 'kwargs_density_profile': {}}
        subhalo_flag = True
        single_subhalo_peak_height = SingleHalo(10**9, x, y, mdef, z, zlens, zsource, r3d, subhalo_flag,
                                                 kwargs_halo_model_2, lens_cosmo=lens_cosmo)
        subhalo_flag = False
        single_fieldhalo_peak_height = SingleHalo(10**10, x, y, mdef, z, zlens, zsource, r3d, subhalo_flag,
                                                 kwargs_halo_model_2, lens_cosmo=lens_cosmo)

        realization = single_subhalo_diemer_joyce.join(single_fieldhalo_diemer_joyce).join(single_subhalo_peak_height).\
            join(single_fieldhalo_peak_height)

        subhalo_diemer_joyce = realization.halos[3]
        fieldhalo_diemer_joyce = realization.halos[2]
        subhalo_peak_height = realization.halos[1]
        fieldhalo_peak_height = realization.halos[0]

        npt.assert_string_equal('DIEMERJOYCE19', subhalo_diemer_joyce._concentration_class.name)
        npt.assert_string_equal('DIEMERJOYCE19', fieldhalo_diemer_joyce._concentration_class.name)
        npt.assert_string_equal('PEAK_HEIGHT_POWERLAW', subhalo_peak_height._concentration_class.name)
        npt.assert_string_equal('PEAK_HEIGHT_POWERLAW', fieldhalo_peak_height._concentration_class.name)
        npt.assert_string_equal('DIEMERJOYCE19', kwargs_halo_model_1['concentration_model'].name)
        npt.assert_string_equal('PEAK_HEIGHT_POWERLAW', kwargs_halo_model_2['concentration_model'].name)

        kwargs_halo_model = {'concentration_model_subhalos': ConcentrationDiemerJoyce(lens_cosmo.cosmo.astropy),
                               'concentration_model_field_halos': ConcentrationPeakHeight(lens_cosmo.cosmo.astropy, 15.0, -0.1, -0.8),
                               'truncation_model_subhalos': TruncationRN(lens_cosmo),
                             'truncation_model_field_halos': TruncationRN(lens_cosmo), 'kwargs_density_profile': {}}
        subhalo_flag = True
        single_subhalo_diemer_joyce = SingleHalo(10 ** 7, x, y, mdef, z, zlens, zsource, r3d, subhalo_flag,
                                                 kwargs_halo_model, lens_cosmo=lens_cosmo)
        subhalo_flag = False
        single_fieldhalo_peak_height = SingleHalo(10 ** 8, x, y, mdef, z, zlens, zsource, r3d, subhalo_flag,
                                                   kwargs_halo_model, lens_cosmo=lens_cosmo)



        realization = single_subhalo_diemer_joyce.join(single_fieldhalo_peak_height)

        subhalo_diemer_joyce = realization.halos[1]
        fieldhalo_peak_height = realization.halos[0]

        npt.assert_string_equal('DIEMERJOYCE19', subhalo_diemer_joyce._concentration_class.name)
        npt.assert_string_equal('PEAK_HEIGHT_POWERLAW', fieldhalo_peak_height._concentration_class.name)
        npt.assert_string_equal('DIEMERJOYCE19', kwargs_halo_model['concentration_model_subhalos'].name)
        npt.assert_string_equal('PEAK_HEIGHT_POWERLAW', kwargs_halo_model['concentration_model_field_halos'].name)

    def test_mass_at_z_exact(self):

        astropy = self.realization.lens_cosmo.cosmo.astropy
        concentration_model = ConcentrationDiemerJoyce(astropy)
        halo_z = 0.3
        z_lens = 0.5
        realization = SingleHalo(10**8, 0.5, 1.0, 'NFW', halo_z, z_lens, 1.5, r3d=None, subhalo_flag=False,
                 kwargs_halo_model={'concentration_model': concentration_model,
                                    'truncation_model': None,
                                    'kwargs_density_profile': {}})
        npt.assert_equal(realization.halos[0].mass, realization.mass_at_z_exact(halo_z))
        npt.assert_equal(0.0, realization.mass_at_z_exact(z_lens))

    def test_number_of_halos_before_after_z(self):

        astropy = self.realization.lens_cosmo.cosmo.astropy
        concentration_model = ConcentrationDiemerJoyce(astropy)
        halo_z = 0.3
        z_lens = 0.5
        halo_1 = SingleHalo(10 ** 8, 0.5, 1.0, 'NFW', halo_z, z_lens, 1.5, r3d=None, subhalo_flag=False,
                                 kwargs_halo_model={'concentration_model': concentration_model,
                                                    'truncation_model': None,
                                                    'kwargs_density_profile': {}})
        halo_z = 0.3
        halo_2 = SingleHalo(10 ** 8, 0.5, 1.0, 'NFW', halo_z, z_lens, 1.5, r3d=None, subhalo_flag=False,
                            kwargs_halo_model={'concentration_model': concentration_model,
                                               'truncation_model': None,
                                               'kwargs_density_profile': {}})
        halo_z = 0.7
        halo_3 = SingleHalo(10 ** 8, 0.5, 1.0, 'NFW', halo_z, z_lens, 1.5, r3d=None, subhalo_flag=False,
                            kwargs_halo_model={'concentration_model': concentration_model,
                                               'truncation_model': None,
                                               'kwargs_density_profile': {}})
        realization = halo_1.join(halo_2).join(halo_3)
        npt.assert_equal(realization.number_of_halos_after_redshift(0.1), 3)
        npt.assert_equal(realization.number_of_halos_after_redshift(z_lens), 1)
        npt.assert_equal(realization.number_of_halos_before_redshift(2.0), 3)
        npt.assert_equal(realization.number_of_halos_at_redshift(0.7), 1)

    def test_plot(self):

        fig = plt.figure(1)
        ax1 = plt.subplot(111, projection='3d')
        self.realization.plot(ax1)

if __name__ == '__main__':
    pytest.main()




import numpy as np
from lenstronomy.LensModel.lens_model import LensModel
from pyHalo.Rendering.rendering_class_base import RenderingClassBase
from pyHalo.Rendering.SpatialDistributions.correlated import Correlated2D
from pyHalo.Rendering.MassFunctions.delta_function import DeltaFunction
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.single_realization import realization_at_z

class CorrelatedStructure(RenderingClassBase):

    """
    This class generates a population of halos with a spatial distribution that tracks the dark matter density in halos
    at each lens plane
    """
    def __init__(self, mass_function_model, kwargs_mass_function,
                 geometry, lens_cosmo, lens_plane_redshifts, delta_z_list, realization):

        self._cylinder_geometry = Geometry(lens_cosmo.cosmo, lens_cosmo.z_lens, lens_cosmo.z_source,
                                           1.0, 'CYLINDER')
        spatial_distribution_model = Correlated2D(self._cylinder_geometry)
        super(CorrelatedStructure, self).__init__(mass_function_model, kwargs_mass_function, spatial_distribution_model,
                 geometry, lens_cosmo, lens_plane_redshifts, delta_z_list)
        self._realization = realization

    def render(self, rmax, x_center_interp, y_center_interp, arcsec_per_pixel):

        """
        Generates halo masses and positions for correlated structure along the line of sight around
        the angular coordinate of each light ray
        :r_max: the maximum radius in arcsec around (x_center, y_center) around which to generate halos
        :param x_center_interp: an interp1d function that returns the x angular position of a
        ray given a comoving distance
        :param y_center_interp: an interp1d function that returns the y angular position of a
        ray given a comoving distance
        :param arcsec_per_pixel: sets the spatial resolution for the rendering of correlated structure
        :return: mass (in Msun), x (arcsec), y (arcsec), r3d (kpc), redshift
        """

        masses = np.array([])
        x = np.array([])
        y = np.array([])
        redshifts = np.array([])

        plane_redshifts = self._realization.unique_redshifts
        delta_z = []
        rescale_inds = []
        rescale_factors = []

        for i, zi in enumerate(plane_redshifts[0:-1]):
            delta_z.append(plane_redshifts[i + 1] - plane_redshifts[i])
        delta_z.append(self._realization.lens_cosmo.z_source - plane_redshifts[-1])

        for z, dz in zip(plane_redshifts, delta_z):

            #rendering_radius = rmax * self._cylinder_geometry.rendering_scale(z)
            d = self._lens_cosmo.cosmo.D_C_transverse(z)
            x_angle = x_center_interp(d)
            y_angle = y_center_interp(d)
            _m, _x, _y, halo_inds, rescale_factor = self.render_at_z(z, x_angle, y_angle,
                                                rmax, arcsec_per_pixel)
            if len(_m) > 0:
                _z = np.array([z] * len(_x))
                masses = np.append(masses, _m)
                x = np.append(x, _x)
                y = np.append(y, _y)
                redshifts = np.append(redshifts, _z)
                rescale_inds += halo_inds
                rescale_factors += [rescale_factor] * len(halo_inds)

        subhalo_flag = [False] * len(masses)
        r3d = np.array([None] * len(masses))

        return masses, x, y, r3d, redshifts, subhalo_flag, rescale_inds, rescale_factors

    def render_at_z(self, z, angular_coordinate_x, angular_coordinate_y, rendering_radius, arcsec_per_pixel):

        """

        :param n: number of objects to render
        :param z: redshift
        :param angular_coordinate_x: the angular coordinate in arcsec of a light ray at redshift z
        :param angular_coordinate_y: the angular coordinate in arcsec of a light ray at redshift z
        :param rendering_radius: the angular radius inside which to render objects
        :param arcsec_per_pixel: sets the spatial resolution for the rendering of correlated structure
        :return: the positions in arcsec of the rendered objects
        """

        kpc_per_asec = self._cylinder_geometry.kpc_per_arcsec(z)

        pdf, mass_in_area, halo_indexes = self._kappa_at_lens_plane(z, angular_coordinate_x, angular_coordinate_y, rendering_radius,
                                                      arcsec_per_pixel)

        if np.sum(pdf) == 0:
            return np.array([]), np.array([]), np.array([]), [], 1.

        m, rescale_factor = self.render_masses_at_z(mass_in_area)

        n_halos = len(m)
        if n_halos > 0:
            x_kpc, y_kpc = self._spatial_distribution_model.draw(n_halos, rendering_radius, pdf, z,
                                                                angular_coordinate_x, angular_coordinate_y)


            x_arcsec = x_kpc / kpc_per_asec
            y_arcsec = y_kpc / kpc_per_asec
            return m, x_arcsec, y_arcsec, halo_indexes, rescale_factor
        else:
            return np.array([]), np.array([]), np.array([]), [], 1.

    def render_masses_at_z(self, mass_in_area):

        """
        :param mass_in_area: total mass at the lens plane
        :return: halo masses at the desired redshift in units Msun, and the factor by which to rescale the
        original halo profiles
        """

        if self._mass_function_model.name == 'DELTA_FUNCTION':
            rescale_factor = 1.-self._kwargs_mass_function['mass_fraction']
            volume = 1.
            rho = self._kwargs_mass_function['mass_fraction'] * mass_in_area
            mass = 10 ** self._kwargs_mass_function['logM']
            mass_function = self._mass_function_model(mass, volume, rho, self._kwargs_mass_function['draw_poisson'])
        else:
            raise Exception('this class is only implemented for a delta function mass function')

        return mass_function.draw(), rescale_factor

    def _kappa_at_lens_plane(self, z, angular_coordinate_x, angular_coordinate_y,
                            rendering_radius, arcsec_per_pixel):

        """
        This routine computes the 2D convergence map of all halos at a certain redshift centered around (angular_coordinate_x,
        angular_coordinate_y) out to a radius "rendering_radius"

        :param z: redshift
        :param angular_coordinate_x: x center in arcsec
        :param angular_coordinate_y: y center in arcsec
        :param rendering_radius: radius in arcsec
        :param arcsec_per_pixel: pixel size in arcsec
        :return: the convergence map, the total mass in the convergence map in the specified area,
        the indexes of halos contributing to the projected mass map
        """
        realization_at_plane, halo_indexes = realization_at_z(self._realization,
                                                   z,
                                                   angular_coordinate_x,
                                                   angular_coordinate_y,
                                                   2*rendering_radius,
                                                   mass_sheet_correction=False)

        lens_model_list, _, kwargs_lens, numerical_interp = realization_at_plane.lensing_quantities(
            add_mass_sheet_correction=False)

        if len(lens_model_list) == 0:
            return np.array([]), np.array([]), []

        lens_model = LensModel(lens_model_list, numerical_alpha_class=numerical_interp)
        npix = int(2 * rendering_radius / arcsec_per_pixel)
        _r = np.linspace(-rendering_radius, rendering_radius, npix)
        xx, yy = np.meshgrid(_r, _r)
        shape0 = xx.shape
        xx, yy = xx.ravel(), yy.ravel()
        rr = np.sqrt(xx ** 2 + yy ** 2)
        inds_zero = np.where(rr > rendering_radius)[0].ravel()
        pdf = lens_model.kappa(xx + angular_coordinate_x, yy + angular_coordinate_y, kwargs_lens)
        pdf[inds_zero] = 0.
        inds_nan = np.where(np.isnan(pdf))
        pdf[inds_nan] = 0.
        npixels = len(inds_zero)
        rendering_radius_mpc = rendering_radius * (0.001 * self._cylinder_geometry.kpc_per_arcsec(z))
        effective_area = np.pi * rendering_radius_mpc ** 2 / npixels
        mass_in_area = self._mass_in_area(pdf, z, effective_area)
        return pdf.reshape(shape0), mass_in_area, halo_indexes

    def _mass_in_area(self, kappa_pdf, z, area):

        """

        :param kappa_pdf: a convergence map at the given redshift
        :param z: redshift
        :param area: the "area" of the convergence map in Mpc^2/pixel
        :return: the total mass contained in the convergence map
        """
        sigma_crit = self._realization.lens_cosmo.get_sigma_crit_lensing(
            z, self._realization.lens_cosmo.z_source)
        mass_in_area = np.sum(kappa_pdf * sigma_crit) * area
        return mass_in_area




from scipy.interpolate import interp1d
from pyHalo.Cosmology.cosmology import Cosmology
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Halos.HaloModels.NFW import NFWSubhhalo, NFWFieldHalo
from pyHalo.Halos.HaloModels.TNFW import TNFWFieldHalo, TNFWSubhalo
from pyHalo.Halos.HaloModels.generalized_nfw import GeneralNFWFieldHalo, GeneralNFWSubhalo
from pyHalo.Halos.HaloModels.powerlaw import PowerLawFieldHalo, PowerLawSubhalo
from pyHalo.Halos.HaloModels.PsuedoJaffe import PJaffeSubhalo
from pyHalo.Halos.HaloModels.PTMass import PTMass
from pyHalo.Halos.HaloModels.ULDM import ULDMFieldHalo, ULDMSubhalo
from pyHalo.Halos.HaloModels.gaussian import Gaussian
import numpy as np
from copy import deepcopy


def realization_at_z(realization, z, angular_coordinate_x=None, angular_coordinate_y=None, max_range=None,
                     mass_sheet_correction=True):
    """
    :param realization: an instance of Realization
    :param z: the redshift where we want to extract halos
    :param angular_coordinate_x: if max_range is specified, will only keep halos within
    max_range of (angular_coordinate_x, angular_coordinate_y)
    :param angular_coordinate_y:
    :param max_range: radius in arcseconds where we want to keep halos. If None, will return a new realization class
     that contains all halos at redshift z contained in the input realization class
    :param mass_sheet_correction: whether or not to include convergence sheet correction in returned realization
    :return: a new instance of Realization, and the indexes of halos that were kept from the original realization
    """

    _halos, _indexes = realization.halos_at_z(z)
    halos = []
    indexes = []

    if max_range is not None:
        for i, halo in enumerate(_halos):
            dx, dy = halo.x - angular_coordinate_x, halo.y - angular_coordinate_y
            dr = (dx ** 2 + dy ** 2) ** 0.5
            if dr < max_range:
                halos.append(halo)
                indexes.append(_indexes[i])
    else:
        halos = _halos
        indexes = _indexes

    centerx, centery = realization.rendering_center
    kwargs_halo_model = realization.kwargs_halo_model
    geometry = realization.geometry
    return Realization.from_halos(halos, realization.lens_cosmo,
                                  kwargs_halo_model,
                                  mass_sheet_correction,
                                  realization.rendering_classes,
                                  centerx, centery, geometry), indexes

class Realization(object):

    """
    This is the main class for storing a population of dark matter halos, both in the main lens plane and along the
    line of sight. This class is created by the main pyhalo module.
    """

    def __init__(self, masses, x, y, r3d, mdefs, z, subhalo_flag, lens_cosmo,
                 halos=None, kwargs_halo_model={}, mass_sheet_correction=True,
                 rendering_classes=None, rendering_center_x=None, rendering_center_y=None,
                 geometry=None):

        """

        This class is the main class that stores information regarding realizations of dark matter halos. It is not
        intended to be created directly by the user. Instances of this class are created through the class
        pyHalo or pyHalo_dynamic.

        :param masses: an array of halo masses (units solar mass)
        :param x: an array of halo x-coordinates (units arcsec)
        :param y: an array of halo y-coordinates (units arcsec)
        :param r2d: an array of halo 2-d distances from lens center (units kpc / (kpc/arsec), or arcsec,
        at halo redshift)
        :param r3d: an array of halo 2-d distances from lens center (units kpc / (kpc/arsec), or arcsec,
        at halo redshift)
        :param mdefs: mass definition of each halo
        :param z: halo redshift
        :param subhalo_flag: whether each halo is a subhalo or a regular halo
        :param lens_cosmo: an instance of LensCosmo (see Halos.lens_cosmo)
        :param halos: a list of halo class instances
        :param kwargs_halo_model: keyword arguments for the halo classes
        :param mass_sheet_correction: whether to apply a mass sheet correction
        :param rendering_classes: a list of rendering class instances
        :param rendering_center_x: an instance of scipy.interp1d that returns an angular position given a comoving distance.
        The angular coordinate defines the center of the rendering volume, and halos will be distributed symmetrically around it.
        The value defaults to 0, but is overridden when the method shift_background_to_source is called
        :param rendering_center_y: same as rendering_center_x, but for the y angular coordinate
        :param geometry: (optional, only relevant is subtract_exact_mass_sheets=True is specified in kwargs_realization)
        an instance of Geometry (pyHalo.Cosmology.geometry) that defines the rendering volume
        """

        self.apply_mass_sheet_correction = mass_sheet_correction
        self.geometry = geometry
        self.lens_cosmo = lens_cosmo
        self._zlens, self._zsource = self.lens_cosmo.z_lens, self.lens_cosmo.z_source
        self.astropy_instance = self.lens_cosmo.cosmo.astropy
        self.halos = []
        self._loaded_models = {}
        self._has_been_shifted = False
        self.kwargs_halo_model = kwargs_halo_model

        if halos is None:
            for mi, xi, yi, r3di, mdefi, zi, sub_flag in zip(masses, x, y, r3d,
                           mdefs, z, subhalo_flag):

                unique_tag = np.random.rand()
                model = self._load_halo_model(mi, xi, yi, r3di, mdefi, zi, sub_flag, self.lens_cosmo,
                                              kwargs_halo_model, unique_tag)
                self.halos.append(model)
        else:
            self.halos = halos
        self._reset()
        self.set_rendering_classes(rendering_classes)
        if rendering_center_x is None or rendering_center_y is None:
            _z = np.linspace(0, self._zsource, 100)
            d = [self.lens_cosmo.cosmo.D_C_transverse(zi) for zi in _z]
            angle = np.zeros_like(d)
            rendering_center_x = interp1d(d, angle)
            rendering_center_y = interp1d(d, angle)
        self._rendering_center_x = rendering_center_x
        self._rendering_center_y = rendering_center_y

    @classmethod
    def from_halos(cls, halos, lens_cosmo, kwargs_halo_model, msheet_correction, rendering_classes,
                   rendering_center_x=None, rendering_center_y=None, geometry=None):

        """

        :param halos: a list of halo class instances
        :param lens_cosmo: an instance of LensCosmo (see Halos.lens_cosmo)
        :param kwargs_halo_model: keyword arguments for the halo models
        :param msheet_correction: whether or not to apply a mass sheet correction
        :param rendering_classes: a list of rendering classes
        :param rendering_center_x: an instance of scipy.interp1d that returns an angular position given a comoving distance.
        The angular coordinate defines the center of the rendering volume, and halos will be distributed symmetrically around it.
        The value defaults to 0, but is overridden when the method shift_background_to_source is called
        :param rendering_center_y: same as rendering_center_x, but for the y angular coordinate
        :param geometry: (optional, only relevant is subtract_exact_mass_sheets=True is specified in kwargs_realization)
        an instance of Geometry (pyHalo.Cosmology.geometry) that defines the rendering volume
        :return: an instance of Realization created directly from the halo class instances
        """

        realization = Realization(None, None, None, None, None, None, None, lens_cosmo,
                                  halos=halos, kwargs_halo_model=kwargs_halo_model,
                                  mass_sheet_correction=msheet_correction,
                                  rendering_classes=rendering_classes,
                                  rendering_center_x=rendering_center_x,
                                  rendering_center_y=rendering_center_y,
                                  geometry=geometry)
        return realization

    @property
    def rendering_center(self):

        """
        Returns the instances of scipy.interp1d that compute the coordinate center of the lensing volume given a comoving
        distance.
        """
        return self._rendering_center_x, self._rendering_center_y

    def filter(self, aperture_radius_front,
               aperture_radius_back,
               log_mass_allowed_in_aperture_front,
               log_mass_allowed_in_aperture_back,
               log_mass_allowed_global_front,
               log_mass_allowed_global_back,
               interpolated_x_angle, interpolated_y_angle,
               zmin=None, zmax=None, aperture_units='ANGLES'):

        """

        :param aperture_radius_front: the radius of a circular window around each light ray where halos are halo kept
        if they are more massive than log_mass_allowed_in_aperture_front (applied for z < z_lens)
        :param aperture_radius_back: the radius of a circular window around each light ray where halos are halo kept
        if they are more massive than log_mass_allowed_in_aperture_back (applied for z < z_lens)
        :param log_mass_allowed_in_aperture_front: the minimum halo mass to be kept inside the tube around each light ray
        in the foreground
        :param log_mass_allowed_in_aperture_back: the minimum halo mass to be kept inside the tube around each light ray
        in the background
        :param log_mass_allowed_global_front: The minimum mass to be kept everywhere in the foreground (if this is smaller
        than log_mass_allowed_in_aperture_front, then the argument aperture_radius_front will have no effect)
        :param log_mass_allowed_global_back: The minimum mass to be kept everywhere in the background (if this is smaller
        than log_mass_allowed_in_aperture_back, then the argument aperture_radius_back will have no effect)
        :param interpolated_x_angle: a list of scipy.interp1d that retuns the x angular position of a ray in
        arcsec given a comoving distance
        :param interpolated_y_angle: a list of scipy.interp1d that retuns the y angular position of a ray in
        arcsec given a comoving distance
        :param zmin: only keep halos at z > zmin
        :param zmax: only keep halos at z < zmax
        :param aperture_units: either 'ANGLES' or 'MPC'

        - If 'ANGLES', then halos are kept inside angular apertures
        around each light ray with size aperture_radius_front/aperture_radius_back.
        - If 'MPC', then halos are kept inside circular apertures with radius
        R = aperture_radius_front/back * mpc_per_arcsec(0.5)
        where D_C(0.5) is the comoving transverse distance at z = 0.5. The unit of R is arcsec * Mpc

        'ANGLES' is more conservative in that it keeps more halos in the lens model; the rendering area is basically a cone
        since the aperture size is a fixed angle at every redshift, whereas 'MPC' distributes halos in cylindrical
        tubes around each light ray along the line of sight.

        :return: A new instance of Realization with the cuts on position and mass applied
        """
        halos = []

        if zmax is None:
            zmax = self._zsource
        if zmin is None:
            zmin = 0

        for plane_index, zi in enumerate(self.unique_redshifts):

            plane_halos, _ = self.halos_at_z(zi)

            inds_at_z = np.where(self.redshifts == zi)[0]
            x_at_z = self.x[inds_at_z]
            y_at_z = self.y[inds_at_z]
            masses_at_z = np.absolute(self.masses[inds_at_z])

            if zi < zmin:
                continue
            if zi > zmax:
                continue

            comoving_distance_z = self.lens_cosmo.cosmo.D_C_z(zi)

            if zi <= self._zlens:

                minimum_mass_everywhere = deepcopy(log_mass_allowed_global_front)
                minimum_mass_in_window = deepcopy(log_mass_allowed_in_aperture_front)
                aperture_radius_arcsec = deepcopy(aperture_radius_front)

            else:

                minimum_mass_everywhere = deepcopy(log_mass_allowed_global_back)
                minimum_mass_in_window = deepcopy(log_mass_allowed_in_aperture_back)
                aperture_radius_arcsec = deepcopy(aperture_radius_back)

            keep_inds_mass = np.where(masses_at_z >= 10 ** minimum_mass_everywhere)[0]
            inds_m_low = np.where(masses_at_z < 10 ** minimum_mass_everywhere)[0]
            keep_inds_dr = []

            for idx in inds_m_low:

                for k, (interp_x, interp_y) in enumerate(zip(interpolated_x_angle, interpolated_y_angle)):

                    dx = x_at_z[idx] - interp_x(comoving_distance_z)
                    dy = y_at_z[idx] - interp_y(comoving_distance_z)

                    if aperture_units == 'ANGLES':
                        dr_cut = aperture_radius_arcsec

                    elif aperture_units == 'MPC':
                        dx *= comoving_distance_z
                        dy *= comoving_distance_z
                        dr_cut = aperture_radius_arcsec * self.lens_cosmo.cosmo.D_C_z(0.5)
                    else:
                        raise Exception('aperture units must be either MPC or ANGLES')

                    dr = np.sqrt(dx ** 2 + dy ** 2)

                    if dr <= dr_cut:
                        keep_inds_dr.append(idx)
                        break

            keep_inds = np.append(keep_inds_mass, np.array(keep_inds_dr)).astype(int)

            tempmasses = np.absolute(masses_at_z[keep_inds])

            keep_inds = keep_inds[np.where(tempmasses >= 10 ** minimum_mass_in_window)[0]]

            for halo_index in keep_inds:
                halos.append(plane_halos[halo_index])

        return Realization.from_halos(halos, self.lens_cosmo, self.kwargs_halo_model,
                                      self.apply_mass_sheet_correction, self.rendering_classes,
                                      self._rendering_center_x, self._rendering_center_y, self.geometry)

    def set_rendering_classes(self, rendering_classes):

        """
        This method sets the rendering classes for the realization, which are used to apply the negative convergence sheet
        corrections after adding halos. The properties of the convergence sheets you need to add depend on the form of the
        mass function you have specified, so this information is stored in the classes used to render halos
        (LOSPowerLaw, MainLensPowerLaw, etc, see classes in pyHalo/Rendering)

        If the rendering classes are not specified for whatever reason, the code will still run but no negative convergence
        sheets will be included in your lens models. This could potentially bias results as you've effectively made every
        light cone overdense relative to the mean matter density in the Universe.

        :param rendering_classes: a list or an instance of a rendering class (LOSPowerLaw, MainLensPowerLaw)
        """

        if not isinstance(rendering_classes, list):
            rendering_classes = [rendering_classes]
        self.rendering_classes = rendering_classes

    def join(self, real, join_rendering_classes=False):

        """
        This routine combines one realization with another realization, keeping only the unqiue halos
        (as identified by their .unique_tag attribute) between them.

        :param real: another realization, possibly a filtered version of self
        :param join_rendering_classes: If True, the rendering classes associated with the new
        realization will include both the rendering class associated with self and that of real.

        :return: a new realization that contains all unique halos from self and real
        """

        halos = []
        tags = self._tags(self.halos)
        real_tags = self._tags(real.halos)
        if len(tags) >= len(real_tags):
            long, short = tags, real_tags
            halos_long, halos_short = self.halos, real.halos
        else:
            long, short = real_tags, tags
            halos_long, halos_short = real.halos, self.halos

        for halo in halos_short:
            halos.append(halo)

        for i, tag in enumerate(long):

            if tag not in short:
                halos.append(halos_long[i])

        if join_rendering_classes:
            rendering_class_self = self.rendering_classes
            rendering_class_new = real.rendering_classes
            rendering_classes = rendering_class_self + rendering_class_new
        else:
            rendering_classes = self.rendering_classes

        centerx, centery = self.rendering_center
        return Realization.from_halos(halos, self.lens_cosmo, self.kwargs_halo_model,
                                      self.apply_mass_sheet_correction, rendering_classes,
                                      centerx, centery, self.geometry)

    def shift_background_to_source(self, ray_interp_x, ray_interp_y, min_shift_redshift=0):

        """
        This routine shifts the entire relation along a path specified by ray_interp_x/y. This routine is intended
        to be used in situations where the source is significantly offset from the origin, and you want to align the
        center of the rendering volume such that tracks the path of the light.

        :param ray_interp_x: instance of scipy.interp1d, returns the angular position of a ray
        fired through the lens center given a comoving distance
        :param ray_interp_y: same but for the y coordinate
        :param min_shift_redshift: only apply shift when halo redshift exceeds min_shift_redshift
        :return:
        """

        if self._has_been_shifted:
            return self

        halos = []

        for halo in self.halos:

            if halo.fixed_position:
                halos.append(halo)
                continue
            if halo.z < min_shift_redshift:
                continue

            comoving_distance_z = self.lens_cosmo.cosmo.D_C_z(halo.z)
            xshift, yshift = ray_interp_x(comoving_distance_z), ray_interp_y(comoving_distance_z)
            halo.x += xshift
            halo.y += yshift
            halos.append(halo)

        new_realization = Realization.from_halos(halos, self.lens_cosmo, self.kwargs_halo_model, self.apply_mass_sheet_correction,
                                                 self.rendering_classes, ray_interp_x, ray_interp_y, self.geometry)

        new_realization._has_been_shifted = True

        return new_realization

    def lensing_quantities(self, add_mass_sheet_correction=True, kwargs_mass_sheet={}):

        """
        :param add_mass_sheet_correction: include sheets of negative convergence to correct for mass added subhalos/field halos
        :param kwargs_mass_sheet: keyword arguments for the mass sheets, see the method _mass_sheet_correction
        :return: the lens_model_list, redshift_list, kwargs_lens, and numerical_alpha_class keywords that can be plugged
        directly into a lenstronomy LensModel class
        """

        kwargs_lens = []
        lens_model_list = []
        redshift_array = []
        numerical_interp = None

        for i, halo in enumerate(self.halos):

            lens_model_name = halo.lenstronomy_ID
            kwargs_halo, interp_class = halo.lenstronomy_params
            lens_model_list += lens_model_name
            kwargs_lens += kwargs_halo
            redshift_array += [halo.z] * len(lens_model_name)

            if interp_class is not None:
                numerical_interp = interp_class

        if self.apply_mass_sheet_correction and add_mass_sheet_correction:
            if self.rendering_classes is None:
                raise Exception('if applying a convergence sheet correction, must specify '
                                'the rendering classes used to determine them.')
            kwargs_mass_sheets, profile_list, z_sheets = self._mass_sheet_correction(self.rendering_classes,
                                                                                     **kwargs_mass_sheet)
            kwargs_lens += kwargs_mass_sheets
            lens_model_list += profile_list
            redshift_array = np.append(redshift_array, z_sheets)

        return lens_model_list, redshift_array, kwargs_lens, numerical_interp

    def split_at_z(self, z):
        """
        Splits the realization at redshift z, returning one instance at Realization containing all halos with
        redshift < zlens and another with all halos at redshift >= z. Be careful with the mass sheet corrections contained
        in the rendering_classes, as both new realizations will get all rendering classes from the parent realization.

        :param z: the redshift at which to split the realization
        :return: two instances at Realization divided at redshift z
        """

        halos_1, halos_2 = [], []
        for halo in self.halos:
            if halo.z <= z:
                halos_1.append(halo)
            else:
                halos_2.append(halo)

        centerx, centery = self.rendering_center
        realization_1 = Realization.from_halos(halos_1, self.lens_cosmo,
                                               self.kwargs_halo_model, self.apply_mass_sheet_correction, self.rendering_classes,
                                               centerx, centery, self.geometry)
        realization_2 = Realization.from_halos(halos_2, self.lens_cosmo,
                                               self.kwargs_halo_model, self.apply_mass_sheet_correction, self.rendering_classes,
                                               centerx, centery, self.geometry)

        return realization_1, realization_2

    def halo_comoving_coordinates(self):

        """
        :param halos: a list of halos
        :return: the comoving (x, y) position, mass, and redshift of each halo in the realization
        """
        xcoords, ycoords, masses, redshifts = [], [], [], []

        for halo in self.halos:
            D = self.lens_cosmo.cosmo.D_C_z(halo.z)
            x_arcsec, y_arcsec = halo.x, halo.y
            x_comoving, y_comoving = D * x_arcsec, D * y_arcsec
            xcoords.append(x_comoving)
            ycoords.append(y_comoving)
            masses.append(halo.mass)
            redshifts.append(halo.z)

        return np.array(xcoords), np.array(ycoords), np.log10(masses), np.array(redshifts)

    def halos_at_z(self, z):
        """

        :param z: redshift
        :return: all halos in the realization that are at redshift z
        """
        halos = []
        index = []
        for i, halo in enumerate(self.halos):
            if halo.z != z:
                continue
            halos.append(halo)
            index.append(i)

        return halos, index

    def mass_at_z_exact(self, z):

        """
        Computes the total mass rendered at each redshift z
        :param z: redshift
        :return: total mass rendered at z
        """

        inds = np.where(self.redshifts == z)
        m_exact = np.sum(self.masses[inds])
        return m_exact

    def number_of_halos_before_redshift(self, z):

        """
        Computes the number of halos with redshifts < z
        :param z: redshift
        :return: number of halos with redshift < z
        """
        n = 0
        for halo in self.halos:
            if halo.z < z:
                n += 1
        return n

    def number_of_halos_after_redshift(self, z):

        """
        Computes the number of halos with redshifts > z
        :param z: redshift
        :return: number of halos with redshift > z
        """
        n = 0
        for halo in self.halos:
            if halo.z > z:
                n += 1
        return n

    def number_of_halos_at_redshift(self, z):

        """
        Computes the number of halos with redshifts == z
        :param z: redshift
        :return: number of halos at z
        """

        n = 0
        for halo in self.halos:
            if halo.z == z:
                n += 1
        return n

    def _mass_sheet_correction(self, rendering_classes, subtract_exact_sheets=False,
                               kappa_scale=1.0, log_mlow_sheets=7.0, log_mhigh_sheets=10.0, zmin=None, zmax=None):

        """
        This routine adds the negative mass sheet corrections along the LOS and in the main lens plane.
        The actual physics that determines the amount of negative convergence to add is encoded in the rendering_classes
        (see for example Rendering.Field.PowerLaw.powerlaw_base.py)

        :param rendering_classes:
        :param subtract_exact_sheets:
        :param kappa_scale:
        :param log_mlow_sheets:
        :param log_mhigh_sheets:
        :param zmin:
        :param zmax:
        :return:
        """

        kwargs_mass_sheets_out = []
        redshifts_out = []
        profiles_out = []

        if subtract_exact_sheets:
            redshifts_out = []
            for zi in self.unique_redshifts:
                redshifts_out.append(zi)
                area = self.geometry.angle_to_physical_area(0.5 * self.geometry.cone_opening_angle, zi)
                kappa_ext = -self.mass_at_z_exact(zi) / self.lens_cosmo.sigma_crit_mass(zi, area)
                kwargs_sheet = {'kappa_ext': kappa_ext}
                kwargs_mass_sheets_out.append(kwargs_sheet)
            profiles_out = ['CONVERGENCE'] * len(kwargs_mass_sheets_out)

        else:

            for rendering_class in rendering_classes:

                if rendering_class is None:
                    continue

                kwargs_new, profiles_new, redshifts_new = \
                    rendering_class.convergence_sheet_correction(kappa_scale, log_mlow_sheets,
                                                                 log_mhigh_sheets, zmin, zmax)

                kwargs_mass_sheets_out += kwargs_new
                redshifts_out += redshifts_new
                profiles_out += profiles_new

        # define the center of mass sheet to be the center of the rendering volume
        centerx_interp, centery_interp = self.rendering_center
        for i, (zi, profile_name) in enumerate(zip(redshifts_out, profiles_out)):
            di = self.lens_cosmo.cosmo.D_C_z(zi)
            x_center, y_center = centerx_interp(di), centery_interp(di)
            if profile_name == 'CONVERGENCE':
                kwargs_mass_sheets_out[i]['ra_0'] = float(x_center)
                kwargs_mass_sheets_out[i]['dec_0'] = float(y_center)
            else:
                kwargs_mass_sheets_out[i]['center_x'] = float(x_center)
                kwargs_mass_sheets_out[i]['center_y'] = float(y_center)

        return kwargs_mass_sheets_out, profiles_out, redshifts_out

    @staticmethod
    def _load_halo_model(mass, x, y, r3d, mdef, z, is_subhalo,
                         lens_cosmo_instance, kwargs_halo_model, unique_tag):

        """
        Loads the halo model for each object based on the mass definition
        :param halo: an instance of Halo
        :return: the specific Halo class corresponding to mass definition mdef
        """

        kwargs_halo = {'mass': mass, 'x': x, 'y': y, 'r3d': r3d,
                       'z': z, 'sub_flag': is_subhalo, 'lens_cosmo_instance': lens_cosmo_instance,
                       'unique_tag': unique_tag}

        if 'truncation_model' in kwargs_halo_model.keys():
            kwargs_halo['truncation_class'] = kwargs_halo_model['truncation_model']
        else:
            if is_subhalo:
                kwargs_halo['truncation_class'] = kwargs_halo_model['truncation_model_subhalos']
            else:
                kwargs_halo['truncation_class'] = kwargs_halo_model['truncation_model_field_halos']
        if 'concentration_model' in kwargs_halo_model.keys():
            kwargs_halo['concentration_class'] = kwargs_halo_model['concentration_model']
        else:
            if is_subhalo:
                kwargs_halo['concentration_class'] = kwargs_halo_model['concentration_model_subhalos']
            else:
                kwargs_halo['concentration_class'] = kwargs_halo_model['concentration_model_field_halos']
        kwargs_halo['args'] = kwargs_halo_model['kwargs_density_profile']
        if mdef == 'NFW':
            if is_subhalo:
                model = NFWSubhhalo
            else:
                model = NFWFieldHalo
        elif mdef == 'TNFW':
            if is_subhalo:
                model = TNFWSubhalo
            else:
                model = TNFWFieldHalo
        elif mdef == 'PT_MASS':
            model = PTMass
        elif mdef == 'PJAFFE':
            model = PJaffeSubhalo
        elif mdef == 'ULDM':
            if is_subhalo:
                model = ULDMSubhalo
            else:
                model = ULDMFieldHalo
        elif mdef == 'GAUSSIAN_KAPPA':
            model = Gaussian
        elif mdef == 'GNFW':
            if is_subhalo:
                model = GeneralNFWSubhalo
            else:
                model = GeneralNFWFieldHalo
        elif mdef == 'SPL_CORE':
            if is_subhalo:
                model = PowerLawSubhalo
            else:
                model = PowerLawFieldHalo
        else:
            raise ValueError('halo profile ' + str(mdef) + ' not recongnized.')

        return model(**kwargs_halo)

    def _tags(self, halos=None):

        """

        :param halos: a list of halos
        :return: the unique tag for each halo in halos; if halos is not specified, returns the unique tag for each
        halo in the realization
        """
        if halos is None:
            halos = self.halos
        tags = []

        for halo in halos:

            tags.append(halo.unique_tag)

        return tags

    def _reset(self):
        """
        Resets all class attributes to the current set of halos contained in the realization
        :return:
        """

        self.x = []
        self.y = []
        self.masses = []
        self.redshifts = []
        self.r3d = []
        self.mdefs = []
        self._halo_tags = []
        self.subhalo_flags = []

        for halo in self.halos:
            self.masses.append(halo.mass)
            self.x.append(halo.x)
            self.y.append(halo.y)
            self.redshifts.append(halo.z)
            self.r3d.append(halo.r3d)
            self.mdefs.append(halo.mdef)
            self._halo_tags.append(halo.unique_tag)
            self.subhalo_flags.append(halo.is_subhalo)

        self.masses = np.array(self.masses)
        self.x = np.array(self.x)
        self.y = np.array(self.y)
        self.r3d = np.array(self.r3d)
        self.redshifts = np.array(self.redshifts)

        self.unique_redshifts = np.sort(np.unique(self.redshifts))

    def __eq__(self, other_reealization):

        """
        Defintes equality between two realizations if they contain the same halos with the same unique tags
        :param other_reealization:
        :return:
        """
        tags = self._tags(self.halos)
        other_tags = other_reealization._tags()
        for tag in other_tags:
            if tag not in tags:
                return False
        else:
            return True

    def plot(self, ax, annotation='', color_normalization=6, marker_size_normalization=7.5,
             view_init_1=30.0, view_init_2=70.0, r_max=None, ray_interp_x_list=None, ray_interp_y_list=None,
             scale_rays=1.0):

        from matplotlib import cm
        import matplotlib.pyplot as plt

        cmap = cm.bone
        distance_calc = self.lens_cosmo.cosmo.D_C_transverse
        zlens = self.lens_cosmo.z_lens
        zsource = self.lens_cosmo.z_source
        x_comoving = []
        y_comoving = []
        redshifts = []
        masses = []

        opening_angle = self.geometry.cone_opening_angle
        angle_zlens = opening_angle * self.geometry.rendering_scale(zlens) / 2.2
        angle_zsrc = opening_angle * self.geometry.rendering_scale(zsource) / 2.2
        if r_max is None:
            if angle_zlens > angle_zsrc:
                r_max = distance_calc(zlens) * angle_zlens / 206265
            else:
                r_max = distance_calc(zsource) * angle_zsrc / 206265

        for zi in self.unique_redshifts:

            d = distance_calc(zi)
            halos_at_redshift, _ = self.halos_at_z(zi)
            for halo in halos_at_redshift:
                x_comoving.append(halo.x * d)
                y_comoving.append(halo.y * d)
                redshifts.append(halo.z)
                masses.append(halo.mass)
        masses = np.array(masses)
        colors = []
        for mi in masses:
            rescaled_mass = max(0.2, (np.log10(mi) - color_normalization) / 4)
            rescaled_mass = min(rescaled_mass, 0.75)
            colors.append(cmap(rescaled_mass))

        distances_from_halo_redshifts = np.array([distance_calc(zi) for zi in redshifts])
        sizes = 24 * (masses / 10**marker_size_normalization) ** 1.

        ax.scatter(distances_from_halo_redshifts, np.array(x_comoving) / 206265,
                   np.array(y_comoving) / 206265,
                   s=sizes,
                   color=colors)

        if ray_interp_x_list is not None:
            assert ray_interp_y_list is not None
            z_array_lines = np.linspace(0.01, zsource, 100)
            d = np.array([distance_calc(zi) for zi in z_array_lines])
            for x_interp,y_interp in zip(ray_interp_x_list, ray_interp_y_list):
                x_rays = scale_rays * d*x_interp(d)/206265
                y_rays = scale_rays * d*y_interp(d)/206265
                ax.plot(d, x_rays, y_rays, color='r',lw=3)

        ax.annotate(annotation, xy=(0.02, 0.045), fontsize=20)
        ax.set_xlim(0., distance_calc(1.025 * zsource))
        zplot = np.arange(0., zsource, 0.25)
        xticks = np.array([distance_calc(zi) for zi in zplot])
        ax.set_xticks(xticks)
        ax.set_xticklabels(zplot, fontsize=16)
        ax.set_xlabel('redshift', fontsize=20, labelpad=35)
        ax.set_ylabel('transverse comoving\ndistance [kpc]', fontsize=18, labelpad=35)

        ax.set_ylim(-r_max, r_max)
        ax.set_zlim(-r_max, r_max)
        yz_ticks = np.round(np.linspace(-1000 * r_max, 1000 * r_max, 5), 0) / 1000
        ax.set_yticks(yz_ticks)
        ax.set_yticklabels(1000 * yz_ticks, fontsize=16)
        ax.set_zticks(yz_ticks)
        ax.set_zticklabels(1000 * yz_ticks, fontsize=16)
        ax.tick_params(axis='y', which='major', pad=10)
        ax.tick_params(axis='z', which='major', pad=10)
        #yzticks = np.array([distance_calc(zi) for zi in zplot])
        ax.view_init(view_init_1, view_init_2)
        plt.xticks(rotation=45)


class SingleHalo(Realization):

    def __init__(self, halo_mass, x, y, mdef, z, zlens, zsource, r3d=None, subhalo_flag=False,
                 kwargs_halo_model={}, astropy_instance=None, lens_cosmo=None):

        """
       Useful for generating a realization with a single or a few
        user-specified halos.
        :param halo_mass: mass of the halo in M_sun
        :param x: halo x coordinate in arcsec
        :param y: halo y coordinate in arcsec
        :param mdef: halo mass definition
        :param z: halo redshift
        :param zlens: main deflector redshift
        :param zsource: source redshift
        :param r3d: three dimensional coordinate of halo inside the host in kpc
        (only relevant for tidally-truncated subhalos, for field halos this can be None)
        :param subhalo_flag: bool, sets whether or not a halo is a subhalo
        :param kwargs_halo_model: keyword arguments for the halo
        :param astropy_instance: an instance of astropy
        :param lens_cosmo: an instance of LensCosmo; if none is provided a new one is generated
        """

        if lens_cosmo is None:
            cosmo = Cosmology(astropy_instance)
            lens_cosmo = LensCosmo(zlens, zsource, cosmo)
        super(SingleHalo, self).__init__([halo_mass], [x], [y],
                                         [r3d], [mdef], [z], [subhalo_flag], lens_cosmo,
                                         kwargs_halo_model=kwargs_halo_model, mass_sheet_correction=False)

